# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Control flow UI widget"""

import os.path
import logging
import uuid
import sip
from math import ceil
from timeit import default_timer as timer
from ui.qt import (Qt, QSize, QTimer, QDir, QUrl, QSizeF, QPainter, QImage,
                   QToolBar, QWidget, QPrinter, QApplication, QHBoxLayout,
                   QLabel, QVBoxLayout, QFileDialog,
                   QDialog, QMenu, QToolButton, QMessageBox, QSvgGenerator,
                   QStackedWidget)
from ui.spacers import ToolBarExpandingSpacer, ToolBarVSpacer
from cdmcfparser import getControlFlowFromMemory
from flowui.vcanvas import VirtualCanvas, formatFlow
from flowui.cflowsettings import getCflowSettings
from flowui.cml import CMLVersion
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.fileutils import isPythonMime
from utils.settings import Settings
from utils.diskvaluesrelay import (getFilePosition, getCollapsedGroups,
                                   setCollapsedGroups)
from .flowuinavbar import ControlFlowNavigationBar
from .flowuisceneview import CFGraphicsView


IDLE_TIMEOUT = 1500

SMART_ZOOM_ALL = 0          # Show everything
SMART_ZOOM_NO_CONTENT = 1   # All the boxes are without a content
SMART_ZOOM_CONTROL_ONLY = 2 # Only scopes (except decorators) and ifs
SMART_ZOOM_CLASS_FUNC = 3   # Only classes and functions


def getSmartZoomDescription(smartZoomLevel):
    """Provides a tooltip for the smart zoom level"""
    if smartZoomLevel == SMART_ZOOM_ALL:
        return 'Smart zoom level: everything is shown'
    if smartZoomLevel == SMART_ZOOM_NO_CONTENT:
        return 'Smart zoom level: everything without a content'
    if smartZoomLevel == SMART_ZOOM_CONTROL_ONLY:
        return 'Smart zoom level: scopes (except decorators) and ifs'
    if smartZoomLevel == SMART_ZOOM_CLASS_FUNC:
        return 'Smart zoom level: only classes and functions'
    return 'Unknown smart zoom level'


def tweakSmartSettings(cflowSettings, smartZoom):
    """Tweaks settings in accordance to the smart level"""
    # Default is to show everything
    if smartZoom == SMART_ZOOM_ALL:
        return cflowSettings
    if smartZoom == SMART_ZOOM_NO_CONTENT:
        cflowSettings.noContent = True
        cflowSettings.noComment = True
        cflowSettings.noDocstring = True
        cflowSettings.minWidth = ceil(float(cflowSettings.minWidth) * 0.66)
        return cflowSettings
    if smartZoom == SMART_ZOOM_CONTROL_ONLY:
        cflowSettings.noContent = True
        cflowSettings.noComment = True
        cflowSettings.noDocstring = True
        cflowSettings.noBlock = True
        cflowSettings.noImport = True
        cflowSettings.noContinue = True
        cflowSettings.noBreak = True
        cflowSettings.noReturn = True
        cflowSettings.noRaise = True
        cflowSettings.noAssert = True
        cflowSettings.noSysExit = True
        cflowSettings.noDecor = True
        cflowSettings.noGroup = True
        cflowSettings.minWidth = ceil(float(cflowSettings.minWidth) * 0.66)
        return cflowSettings
    if smartZoom == SMART_ZOOM_CLASS_FUNC:
        cflowSettings.noContent = True
        cflowSettings.noComment = True
        cflowSettings.noDocstring = True
        cflowSettings.noBlock = True
        cflowSettings.noImport = True
        cflowSettings.noContinue = True
        cflowSettings.noBreak = True
        cflowSettings.noReturn = True
        cflowSettings.noRaise = True
        cflowSettings.noAssert = True
        cflowSettings.noSysExit = True
        cflowSettings.noDecor = True
        cflowSettings.noFor = True
        cflowSettings.noWhile = True
        cflowSettings.noWith = True
        cflowSettings.noTry = True
        cflowSettings.noIf = True
        cflowSettings.noGroup = True
        cflowSettings.minWidth = ceil(float(cflowSettings.minWidth) * 0.66)
        return cflowSettings
    return cflowSettings


class FlowUIWidget(QWidget):

    """The widget which goes along with the text editor"""

    def __init__(self, editor, parent):
        QWidget.__init__(self, parent)

        # It is always not visible at the beginning because there is no
        # editor content at the start
        self.setVisible(False)

        self.__editor = editor
        self.__parentWidget = parent
        self.__connected = False
        self.__needPathUpdate = False

        self.cflowSettings = getCflowSettings(self)
        self.__displayProps = (self.cflowSettings.hidedocstrings,
                               self.cflowSettings.hidecomments,
                               self.cflowSettings.hideexcepts,
                               Settings()['smartZoom'])

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)

        vLayout = QVBoxLayout()
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)

        # Make pylint happy
        self.__toolbar = None
        self.__navBar = None
        self.__cf = None
        self.__canvas = None
        self.__validGroups = []
        self.__allGroupId = set()

        # Create the update timer
        self.__updateTimer = QTimer(self)
        self.__updateTimer.setSingleShot(True)
        self.__updateTimer.timeout.connect(self.process)

        vLayout.addWidget(self.__createNavigationBar())
        vLayout.addWidget(self.__createStackedViews())

        hLayout.addLayout(vLayout)
        hLayout.addWidget(self.__createToolbar())
        self.setLayout(hLayout)

        self.updateSettings()

        # Connect to the change file type signal
        self.__mainWindow = GlobalData().mainWindow
        editorsManager = self.__mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.connect(self.__onFileTypeChanged)

        Settings().sigHideDocstringsChanged.connect(
            self.__onHideDocstringsChanged)
        Settings().sigHideCommentsChanged.connect(self.__onHideCommentsChanged)
        Settings().sigHideExceptsChanged.connect(self.__onHideExceptsChanged)
        Settings().sigSmartZoomChanged.connect(self.__onSmartZoomChanged)

        self.setSmartZoomLevel(Settings()['smartZoom'])

    def getParentWidget(self):
        return self.__parentWidget

    def view(self):
        """Provides a reference to the current view"""
        return self.smartViews.currentWidget()

    def scene(self):
        """Provides a reference to the current scene"""
        return self.view().scene

    def __createToolbar(self):
        """Creates the toolbar"""
        self.__toolbar = QToolBar(self)
        self.__toolbar.setOrientation(Qt.Vertical)
        self.__toolbar.setMovable(False)
        self.__toolbar.setAllowedAreas(Qt.RightToolBarArea)
        self.__toolbar.setIconSize(QSize(16, 16))
        self.__toolbar.setFixedWidth(30)
        self.__toolbar.setContentsMargins(0, 0, 0, 0)

        # Buttons
        saveAsMenu = QMenu(self)
        saveAsSVGAct = saveAsMenu.addAction(getIcon('filesvg.png'),
                                            'Save as SVG...')
        saveAsSVGAct.triggered.connect(self.onSaveAsSVG)

        saveAsPDFAct = saveAsMenu.addAction(getIcon('filepdf.png'),
                                            'Save as PDF...')
        saveAsPDFAct.triggered.connect(self.onSaveAsPDF)
        saveAsPNGAct = saveAsMenu.addAction(getIcon('filepixmap.png'),
                                            'Save as PNG...')
        saveAsPNGAct.triggered.connect(self.onSaveAsPNG)
        saveAsMenu.addSeparator()
        saveAsCopyToClipboardAct = saveAsMenu.addAction(
            getIcon('copymenu.png'), 'Copy to clipboard')
        saveAsCopyToClipboardAct.triggered.connect(self.copyToClipboard)

        self.__saveAsButton = QToolButton(self.__toolbar)
        self.__saveAsButton.setIcon(getIcon('saveasmenu.png'))
        self.__saveAsButton.setToolTip('Save as')
        self.__saveAsButton.setPopupMode(QToolButton.InstantPopup)
        self.__saveAsButton.setMenu(saveAsMenu)
        self.__saveAsButton.setFocusPolicy(Qt.NoFocus)

        self.__levelUpButton = QToolButton(self.__toolbar)
        self.__levelUpButton.setFocusPolicy(Qt.NoFocus)
        self.__levelUpButton.setIcon(getIcon('levelup.png'))
        self.__levelUpButton.setToolTip('Smart zoom level up (Shift+wheel)')
        self.__levelUpButton.clicked.connect(self.onSmartZoomLevelUp)
        self.__levelIndicator = QLabel('<b>0</b>', self.__toolbar)
        self.__levelIndicator.setAlignment(Qt.AlignCenter)
        self.__levelIndicator.setStyleSheet('QLabel {background: transparent}')
        self.__levelDownButton = QToolButton(self.__toolbar)
        self.__levelDownButton.setFocusPolicy(Qt.NoFocus)
        self.__levelDownButton.setIcon(getIcon('leveldown.png'))
        self.__levelDownButton.setToolTip('Smart zoom level down (Shift+wheel)')
        self.__levelDownButton.clicked.connect(self.onSmartZoomLevelDown)

        self.__hideDocstrings = QToolButton(self.__toolbar)
        self.__hideDocstrings.setCheckable(True)
        self.__hideDocstrings.setIcon(getIcon('hidedocstrings.png'))
        self.__hideDocstrings.setToolTip('Show/hide docstrings')
        self.__hideDocstrings.setFocusPolicy(Qt.NoFocus)
        self.__hideDocstrings.setChecked(Settings()['hidedocstrings'])
        self.__hideDocstrings.clicked.connect(self.__onHideDocstrings)
        self.__hideComments = QToolButton(self.__toolbar)
        self.__hideComments.setCheckable(True)
        self.__hideComments.setIcon(getIcon('hidecomments.png'))
        self.__hideComments.setToolTip('Show/hide comments')
        self.__hideComments.setFocusPolicy(Qt.NoFocus)
        self.__hideComments.setChecked(Settings()['hidecomments'])
        self.__hideComments.clicked.connect(self.__onHideComments)
        self.__hideExcepts = QToolButton(self.__toolbar)
        self.__hideExcepts.setCheckable(True)
        self.__hideExcepts.setIcon(getIcon('hideexcepts.png'))
        self.__hideExcepts.setToolTip('Show/hide except blocks')
        self.__hideExcepts.setFocusPolicy(Qt.NoFocus)
        self.__hideExcepts.setChecked(Settings()['hideexcepts'])
        self.__hideExcepts.clicked.connect(self.__onHideExcepts)

        self.__toolbar.addWidget(self.__saveAsButton)
        self.__toolbar.addWidget(ToolBarExpandingSpacer(self.__toolbar))
        self.__toolbar.addWidget(self.__levelUpButton)
        self.__toolbar.addWidget(self.__levelIndicator)
        self.__toolbar.addWidget(self.__levelDownButton)
        self.__toolbar.addWidget(ToolBarVSpacer(self.__toolbar, 10))
        self.__toolbar.addWidget(self.__hideDocstrings)
        self.__toolbar.addWidget(self.__hideComments)
        self.__toolbar.addWidget(self.__hideExcepts)

        return self.__toolbar

    def __createNavigationBar(self):
        """Creates the navigation bar"""
        self.__navBar = ControlFlowNavigationBar(self)
        return self.__navBar

    def __createStackedViews(self):
        """Creates the graphics view"""
        self.smartViews = QStackedWidget(self)
        self.smartViews.setContentsMargins(0, 0, 0, 0)

        self.smartViews.addWidget(CFGraphicsView(self.__navBar, self))
        self.smartViews.addWidget(CFGraphicsView(self.__navBar, self))
        return self.smartViews

    def process(self):
        """Parses the content and displays the results"""
        if not self.__connected:
            self.__connectEditorSignals()

        start = timer()
        cf = getControlFlowFromMemory(self.__editor.text)
        end = timer()
        if cf.errors:
            self.__navBar.updateInfoIcon(self.__navBar.STATE_BROKEN_UTD)
            errors = []
            for err in cf.errors:
                if err[0] == -1 and err[1] == -1:
                    errors.append(err[2])
                elif err[1] == -1:
                    errors.append('[' + str(err[0]) + ':] ' + err[2])
                elif err[0] == -1:
                    errors.append('[:' + str(err[1]) + '] ' + err[2])
                else:
                    errors.append('[' + str(err[0]) + ':' +
                                  str(err[1]) + '] ' + err[2])
            self.__navBar.setErrors(errors)
            return

        self.__cf = cf
        if self.isDebugMode():
            logging.info('Parsed file: %s', formatFlow(str(self.__cf)))
            logging.info('Parse timing: %f', end - start)

        # Collect warnings (parser + CML warnings) and valid groups
        self.__validGroups = []
        self.__allGroupId = set()
        allWarnings = self.__cf.warnings + \
                      CMLVersion.validateCMLComments(self.__cf,
                                                     self.__validGroups,
                                                     self.__allGroupId)

        # That will clear the error tooltip as well
        self.__navBar.updateInfoIcon(self.__navBar.STATE_OK_UTD)

        if allWarnings:
            warnings = []
            for warn in allWarnings:
                if warn[0] == -1 and warn[1] == -1:
                    warnings.append(warn[2])
                elif warn[1] == -1:
                    warnings.append('[' + str(warn[0]) + ':] ' + warn[2])
                elif warn[0] == -1:
                    warnings.append('[:' + str(warn[1]) + '] ' + warn[2])
                else:
                    warnings.append('[' + str(warn[0]) + ':' +
                                    str(warn[1]) + '] ' + warn[2])
            self.__navBar.setWarnings(warnings)
        else:
            self.__navBar.clearWarnings()

        self.redrawScene()

    def __cleanupCanvas(self):
        """Cleans up the canvas"""
        if self.__canvas is not None:
            self.__canvas.cleanup()
            self.__canvas = None
        for item in self.scene().items():
            sip.delete(item)
        self.scene().clear()

    def redrawScene(self):
        """Redraws the scene"""
        smartZoomLevel = Settings()['smartZoom']
        self.cflowSettings = getCflowSettings(self)
        if self.dirty():
            self.__displayProps = (self.cflowSettings.hidedocstrings,
                                   self.cflowSettings.hidecomments,
                                   self.cflowSettings.hideexcepts,
                                   smartZoomLevel)
        self.cflowSettings.itemID = 0
        self.cflowSettings = tweakSmartSettings(self.cflowSettings,
                                                smartZoomLevel)

        try:
            fileName = self.__parentWidget.getFileName()
            if not fileName:
                fileName = self.__parentWidget.getShortName()
            collapsedGroups = getCollapsedGroups(fileName)

            # Top level canvas has no adress and no parent canvas
            self.__cleanupCanvas()
            self.__canvas = VirtualCanvas(self.cflowSettings, None, None,
                                          self.__validGroups, collapsedGroups,
                                          None)
            lStart = timer()
            self.__canvas.layoutModule(self.__cf)
            lEnd = timer()
            self.__canvas.setEditor(self.__editor)
            width, height = self.__canvas.render()
            rEnd = timer()
            self.scene().setSceneRect(0, 0, width, height)
            self.__canvas.draw(self.scene(), 0, 0)
            dEnd = timer()
            if self.isDebugMode():
                logging.info('Redrawing is done. Size: %d x %d', width, height)
                logging.info('Layout timing: %f', lEnd - lStart)
                logging.info('Render timing: %f', rEnd - lEnd)
                logging.info('Draw timing: %f', dEnd - rEnd)
        except Exception as exc:
            logging.error(str(exc))
            raise

    def onFlowZoomChanged(self):
        """Triggered when a flow zoom is changed"""
        if self.__cf:
            selection = self.scene().serializeSelection()
            firstOnScreen = self.scene().getFirstLogicalItem()
            self.cflowSettings.onFlowZoomChanged()
            self.redrawScene()
            self.updateNavigationToolbar('')
            self.scene().restoreSelectionByID(selection)
            self.__restoreScroll(firstOnScreen)

    def __onFileTypeChanged(self, fileName, uuid, newFileType):
        """Triggered when a buffer content type has changed"""
        if self.__parentWidget.getUUID() != uuid:
            return

        if not isPythonMime(newFileType):
            self.__disconnectEditorSignals()
            self.__updateTimer.stop()
            self.__cleanupCanvas()
            self.__cf = None
            self.__validGroups = []
            self.setVisible(False)
            self.__navBar.updateInfoIcon(self.__navBar.STATE_UNKNOWN)
            return

        # Update the bar and show it
        self.setVisible(True)
        self.process()

        # The buffer type change event comes when the content is loaded first
        # time. So this is a good point to restore the position
        _, _, _, cflowHPos, cflowVPos = getFilePosition(fileName)
        self.setScrollbarPositions(cflowHPos, cflowVPos)

    def terminate(self):
        """Called when a tab is closed"""
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
        self.__updateTimer.deleteLater()

        self.__disconnectEditorSignals()

        self.__mainWindow = GlobalData().mainWindow
        editorsManager = self.__mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)

        Settings().sigHideDocstringsChanged.disconnect(
            self.__onHideDocstringsChanged)
        Settings().sigHideCommentsChanged.disconnect(self.__onHideCommentsChanged)
        Settings().sigHideExceptsChanged.disconnect(self.__onHideExceptsChanged)
        Settings().sigSmartZoomChanged.disconnect(self.__onSmartZoomChanged)

        # Helps GC to collect more
        self.__cleanupCanvas()
        for index in range(self.smartViews.count()):
            self.smartViews.widget(index).terminate()
            self.smartViews.widget(index).deleteLater()

        self.smartViews.deleteLater()
        self.__navBar.deleteLater()
        self.__cf = None

        self.__saveAsButton.menu().deleteLater()
        self.__saveAsButton.deleteLater()

        self.__levelUpButton.clicked.disconnect(self.onSmartZoomLevelUp)
        self.__levelUpButton.deleteLater()

        self.__levelDownButton.clicked.disconnect(self.onSmartZoomLevelDown)
        self.__levelDownButton.deleteLater()

        self.__hideDocstrings.clicked.disconnect(self.__onHideDocstrings)
        self.__hideDocstrings.deleteLater()

        self.__hideComments.clicked.disconnect(self.__onHideComments)
        self.__hideComments.deleteLater()

        self.__hideExcepts.clicked.disconnect(self.__onHideExcepts)
        self.__hideExcepts.deleteLater()

        self.__toolbar.deleteLater()

        self.__editor = None
        self.__parentWidget = None
        self.cflowSettings = None
        self.__displayProps = None

    def __connectEditorSignals(self):
        """When it is a python file - connect to the editor signals"""
        if not self.__connected:
            self.__editor.cursorPositionChanged.connect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.connect(self.__onBufferChanged)
            self.__connected = True

    def __disconnectEditorSignals(self):
        """Disconnect the editor signals when the file is not a python one"""
        if self.__connected:
            self.__editor.cursorPositionChanged.disconnect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.disconnect(self.__onBufferChanged)
            self.__connected = False

    def __cursorPositionChanged(self):
        """Cursor position changed"""
        # The timer should be reset only in case if the redrawing was delayed
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
            self.__updateTimer.start(IDLE_TIMEOUT)

    def __onBufferChanged(self):
        """Triggered to update status icon and to restart the timer"""
        self.__updateTimer.stop()
        if self.__navBar.getCurrentState() in [self.__navBar.STATE_OK_UTD,
                                               self.__navBar.STATE_OK_CHN,
                                               self.__navBar.STATE_UNKNOWN]:
            self.__navBar.updateInfoIcon(self.__navBar.STATE_OK_CHN)
        else:
            self.__navBar.updateInfoIcon(self.__navBar.STATE_BROKEN_CHN)
        self.__updateTimer.start(IDLE_TIMEOUT)

    def redrawNow(self):
        """Redraw the diagram regardless of the timer"""
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
        self.process()

    def generateNewGroupId(self):
        """Generates a new group ID (string)"""
        # It can also consider the current set of the groups: valid + invalid
        # and generate an integer id which is shorter
        for vacantGroupId in range(1000):
            groupId = str(vacantGroupId)
            if not groupId in self.__allGroupId:
                return groupId
        # Last resort
        return str(uuid.uuid1())

    def updateNavigationToolbar(self, text):
        """Updates the toolbar text"""
        if self.__needPathUpdate:
            self.__navBar.setPath(text)

    def updateSettings(self):
        """Updates settings"""
        self.__needPathUpdate = Settings()['showCFNavigationBar']
        self.__navBar.setPathVisible(self.__needPathUpdate)
        self.__navBar.setPath('')

    def highlightAtAbsPos(self, absPos, line, pos):
        """Scrolls the view to the item closest to absPos and selects it.

        line and pos are 1-based
        """
        item, _ = self.scene().getNearestItem(absPos, line, pos)
        if item:
            GlobalData().mainWindow.setFocusToFloatingRenderer()
            self.scene().clearSelection()
            item.setSelected(True)
            self.view().scrollTo(item)
            self.setFocus()

    def setFocus(self):
        """Sets the focus"""
        self.view().setFocus()

    @staticmethod
    def __getDefaultSaveDir():
        """Provides the default directory to save files to"""
        project = GlobalData().project
        if project.isLoaded():
            return project.getProjectDir()
        return QDir.currentPath()

    def __selectFile(self, extension):
        """Picks a file of a certain extension"""
        dialog = QFileDialog(self, 'Save flowchart as')
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setLabelText(QFileDialog.Accept, "Save")
        dialog.setNameFilter(extension.upper() + " files (*." +
                             extension.lower() + ")")
        urls = []
        for dname in QDir.drives():
            urls.append(QUrl.fromLocalFile(dname.absoluteFilePath()))
        urls.append(QUrl.fromLocalFile(QDir.homePath()))
        project = GlobalData().project
        if project.isLoaded():
            urls.append(QUrl.fromLocalFile(project.getProjectDir()))
        dialog.setSidebarUrls(urls)

        suggestedFName = self.__parentWidget.getFileName()
        if '.' in suggestedFName:
            dotIndex = suggestedFName.rindex('.')
            suggestedFName = suggestedFName[:dotIndex]

        dialog.setDirectory(self.__getDefaultSaveDir())
        dialog.selectFile(suggestedFName + "." + extension.lower())
        dialog.setOption(QFileDialog.DontConfirmOverwrite, False)
        dialog.setOption(QFileDialog.DontUseNativeDialog, True)
        if dialog.exec_() != QDialog.Accepted:
            return None

        fileNames = dialog.selectedFiles()
        fileName = os.path.abspath(str(fileNames[0]))
        if os.path.isdir(fileName):
            logging.error("A file must be selected")
            return None

        if "." not in fileName:
            fileName += "." + extension.lower()

        # Check permissions to write into the file or to a directory
        if os.path.exists(fileName):
            # Check write permissions for the file
            if not os.access(fileName, os.W_OK):
                logging.error("There is no write permissions for " + fileName)
                return None
        else:
            # Check write permissions to the directory
            dirName = os.path.dirname(fileName)
            if not os.access(dirName, os.W_OK):
                logging.error("There is no write permissions for the "
                              "directory " + dirName)
                return None

        if os.path.exists(fileName):
            res = QMessageBox.warning(
                self, "Save flowchart as",
                "<p>The file <b>" + fileName + "</b> already exists.</p>",
                QMessageBox.StandardButtons(QMessageBox.Abort |
                                            QMessageBox.Save),
                QMessageBox.Abort)
            if res == QMessageBox.Abort or res == QMessageBox.Cancel:
                return None

        # All prerequisites are checked, return a file name
        return fileName

    def onSaveAsSVG(self):
        """Triggered on the 'Save as SVG' button"""
        fileName = self.__selectFile("svg")
        if fileName is None:
            return False

        try:
            self.__saveAsSVG(fileName)
        except Exception as excpt:
            logging.error(str(excpt))
            return False
        return True

    def __saveAsSVG(self, fileName):
        """Saves the flowchart as an SVG file"""
        generator = QSvgGenerator()
        generator.setFileName(fileName)
        generator.setSize(QSize(self.scene().width(), self.scene().height()))
        painter = QPainter(generator)
        self.scene().render(painter)
        painter.end()

    def onSaveAsPDF(self):
        """Triggered on the 'Save as PDF' button"""
        fileName = self.__selectFile("pdf")
        if fileName is None:
            return False

        try:
            self.__saveAsPDF(fileName)
        except Exception as excpt:
            logging.error(str(excpt))
            return False
        return True

    def __saveAsPDF(self, fileName):
        """Saves the flowchart as an PDF file"""
        printer = QPrinter()
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setPaperSize(QSizeF(self.scene().width(),
                                    self.scene().height()), QPrinter.Point)
        printer.setFullPage(True)
        printer.setOutputFileName(fileName)

        painter = QPainter(printer)
        self.scene().render(painter)
        painter.end()

    def onSaveAsPNG(self):
        """Triggered on the 'Save as PNG' button"""
        fileName = self.__selectFile("png")
        if fileName is None:
            return False

        try:
            self.__saveAsPNG(fileName)
        except Exception as excpt:
            logging.error(str(excpt))
            return False
        return True

    def __getPNG(self):
        """Renders the scene as PNG"""
        image = QImage(self.scene().width(), self.scene().height(),
                       QImage.Format_ARGB32_Premultiplied)
        painter = QPainter(image)
        # It seems that the better results are without antialiasing
        # painter.setRenderHint( QPainter.Antialiasing )
        self.scene().render(painter)
        painter.end()
        return image

    def __saveAsPNG(self, fileName):
        """Saves the flowchart as an PNG file"""
        image = self.__getPNG()
        image.save(fileName, "PNG")

    def copyToClipboard(self):
        """Copies the rendered scene to the clipboard as an image"""
        image = self.__getPNG()
        clip = QApplication.clipboard()
        clip.setImage(image)

    def getScrollbarPositions(self):
        """Provides the scrollbar positions"""
        hScrollBar = self.view().horizontalScrollBar()
        vScrollBar = self.view().verticalScrollBar()
        return hScrollBar.value(), vScrollBar.value()

    def setScrollbarPositions(self, hPos, vPos):
        """Sets the scrollbar positions for the view"""
        self.view().horizontalScrollBar().setValue(hPos)
        self.view().verticalScrollBar().setValue(vPos)

    def __onHideDocstrings(self):
        """Triggered when a hide docstring button is pressed"""
        Settings()['hidedocstrings'] = not Settings()['hidedocstrings']

    def __onHideDocstringsChanged(self):
        """Signalled by settings"""
        selection = self.scene().serializeSelection()
        firstOnScreen = self.scene().getFirstLogicalItem()
        settings = Settings()
        self.__hideDocstrings.setChecked(settings['hidedocstrings'])
        if self.__checkNeedRedraw():
            self.scene().restoreSelectionByID(selection)
            self.__restoreScroll(firstOnScreen)

    def __onHideComments(self):
        """Triggered when a hide comments button is pressed"""
        Settings()['hidecomments'] = not Settings()['hidecomments']

    def __onHideCommentsChanged(self):
        """Signalled by settings"""
        selection = self.scene().serializeSelection()
        firstOnScreen = self.scene().getFirstLogicalItem()
        settings = Settings()
        self.__hideComments.setChecked(settings['hidecomments'])
        if self.__checkNeedRedraw():
            self.scene().restoreSelectionByID(selection)
            self.__restoreScroll(firstOnScreen)

    def __onHideExcepts(self):
        """Triggered when a hide except blocks button is pressed"""
        Settings()['hideexcepts'] = not Settings()['hideexcepts']

    def __onHideExceptsChanged(self):
        """Signalled by settings"""
        selection = self.scene().serializeSelection()
        firstOnScreen = self.scene().getFirstLogicalItem()
        settings = Settings()
        self.__hideExcepts.setChecked(settings['hideexcepts'])
        if self.__checkNeedRedraw():
            self.scene().restoreSelectionByTooltip(selection)
            self.__restoreScroll(firstOnScreen)

    def __checkNeedRedraw(self):
        """Redraws the scene if necessary when a display setting is changed"""
        editorsManager = self.__mainWindow.editorsManagerWidget.editorsManager
        if self.__parentWidget == editorsManager.currentWidget():
            self.updateNavigationToolbar('')
            self.process()
            return True
        return False

    def dirty(self):
        """True if some other tab has switched display settings"""
        settings = Settings()
        return self.__displayProps[0] != settings['hidedocstrings'] or \
            self.__displayProps[1] != settings['hidecomments'] or \
            self.__displayProps[2] != settings['hideexcepts'] or \
            self.__displayProps[3] != settings['smartZoom']

    def onSmartZoomLevelUp(self):
        """Triggered when an upper smart zoom level was requested"""
        Settings().onSmartZoomIn()

    def onSmartZoomLevelDown(self):
        """Triggered when an lower smart zoom level was requested"""
        Settings().onSmartZoomOut()

    def setSmartZoomLevel(self, smartZoomLevel):
        """Sets the new smart zoom level"""
        maxSmartZoom = Settings().MAX_SMART_ZOOM
        if smartZoomLevel < 0 or smartZoomLevel > maxSmartZoom:
            return

        self.__levelIndicator.setText('<b>' + str(smartZoomLevel) + '</b>')
        self.__levelIndicator.setToolTip(
            getSmartZoomDescription(smartZoomLevel))
        self.__levelUpButton.setEnabled(smartZoomLevel < maxSmartZoom)
        self.__levelDownButton.setEnabled(smartZoomLevel > 0)
        self.smartViews.setCurrentIndex(smartZoomLevel)

    def __onSmartZoomChanged(self):
        """Triggered when a smart zoom changed"""
        selection = self.scene().serializeSelection()
        firstOnScreen = self.scene().getFirstLogicalItem()
        self.setSmartZoomLevel(Settings()['smartZoom'])
        if self.__checkNeedRedraw():
            self.scene().restoreSelectionByTooltip(selection)
            self.__restoreScroll(firstOnScreen)

    def __restoreScroll(self, toItem):
        """Restores the view scrolling to the best possible position"""
        if toItem:
            lineRange = toItem.getLineRange()
            absPosRange = toItem.getAbsPosRange()
            item, _ = self.scene().getNearestItem(absPosRange[0],
                                                  lineRange[0], 0)
            if item:
                self.view().scrollTo(item, True)
                self.view().horizontalScrollBar().setValue(0)

    def validateCollapsedGroups(self, fileName):
        """Checks that there are no collapsed groups which are invalid"""
        if self.__navBar.getCurrentState() != self.__navBar.STATE_OK_UTD:
            return

        collapsedGroups = getCollapsedGroups(fileName)
        if collapsedGroups:
            toBeDeleted = []
            for groupId in collapsedGroups:
                for validId, start, end in self.__validGroups:
                    del start
                    del end
                    if validId == groupId:
                        break
                else:
                    toBeDeleted.append(groupId)

            if toBeDeleted:
                for groupId in toBeDeleted:
                    collapsedGroups.remove(groupId)
                setCollapsedGroups(fileName, collapsedGroups)
        else:
            setCollapsedGroups(fileName, [])

    def getDocItemByAnchor(self, anchor):
        """Provides the graphics item for the given anchor if so"""
        return self.scene().getDocItemByAnchor(anchor)

    @staticmethod
    def isDebugMode():
        """True if it is a debug mode"""
        return GlobalData().skin['debug']



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Thread viewer"""


from ui.qt import (Qt, QTreeWidget, QToolButton, QTreeWidgetItem,
                   QHeaderView, QVBoxLayout, QWidget, QSize, QToolBar,
                   QAbstractItemView, QSizePolicy)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from utils.pixmapcache import getIcon
from utils.settings import Settings


class ThreadItem(QTreeWidgetItem):

    """Single thread item data structure"""

    def __init__(self, tid, name, state):
        QTreeWidgetItem.__init__(self, ['', name, state, str(tid)])

        self.__isCurrent = False
        self.__setTooltip()

    def __setTooltip(self):
        """Sets the tooltip"""
        tooltip = "Current: " + str(self.__isCurrent) + "\n" \
                  "Name: " + self.getName() + "\n" \
                  "State: " + self.getState() + "\n" \
                  "TID: " + str(self.getTID())
        for index in range(4):
            self.setToolTip(index, tooltip)

    def setCurrent(self, value):
        """Mark the current thread with an icon if so"""
        if self.__isCurrent == value:
            return

        self.__isCurrent = value
        if value:
            self.setIcon(0, getIcon('currentthread.png'))
        else:
            self.setIcon(0, getIcon('empty.png'))
        self.__setTooltip()

    def getTID(self):
        """Provides the thread ID"""
        return int(self.text(3))

    def getName(self):
        """Provides the thread name"""
        return str(self.text(1))

    def getState(self):
        """Provides the thread state"""
        return str(self.text(2))

    def isCurrent(self):
        """True if the project is current"""
        return self.__isCurrent


class ThreadsViewer(QWidget):

    """Implements the threads viewer for a debugger"""

    def __init__(self, debugger, parent=None):
        QWidget.__init__(self, parent)

        self.__debugger = debugger
        self.__createLayout()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__threadsLabel = HeaderFitLabel(self)
        self.__threadsLabel.setText('Threads')
        self.__threadsLabel.setSizePolicy(QSizePolicy.Expanding,
                                          QSizePolicy.Fixed)
        self.__threadsLabel.setMinimumWidth(10)

        self.__showHideButton = QToolButton()
        self.__showHideButton.setAutoRaise(True)
        self.__showHideButton.setIcon(getIcon('less.png'))
        self.__showHideButton.setFixedSize(self.__threadsLabel.height(),
                                           self.__threadsLabel.height())
        self.__showHideButton.setToolTip('Hide threads list')
        self.__showHideButton.setFocusPolicy(Qt.NoFocus)
        self.__showHideButton.clicked.connect(self.__onShowHide)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(16, 16))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__threadsLabel)
        self.headerToolbar.addWidget(self.__showHideButton)

        self.__threadsList = QTreeWidget()
        self.__threadsList.setSortingEnabled(False)
        # I might not need that because of two reasons:
        # - the window has no focus
        # - the window has custom current indicator
        # self.__threadsList.setAlternatingRowColors( True )
        self.__threadsList.setRootIsDecorated(False)
        self.__threadsList.setItemsExpandable(False)
        self.__threadsList.setUniformRowHeights(True)
        self.__threadsList.setSelectionMode(QAbstractItemView.NoSelection)
        self.__threadsList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__threadsList.setItemDelegate(NoOutlineHeightDelegate(4))
        self.__threadsList.setFocusPolicy(Qt.NoFocus)

        self.__threadsList.itemClicked.connect(self.__onThreadClicked)
        self.__threadsList.setHeaderLabels(["", "Name", "State", "TID"])

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.__threadsList)

    def __onShowHide(self, startup=False):
        """Triggered when show/hide button is clicked"""
        if startup or self.__threadsList.isVisible():
            self.__minH = self.minimumHeight()
            self.__maxH = self.maximumHeight()
            self.splitterSize = self.parent().sizes()[1]

            self.__threadsList.setVisible(False)
            self.__showHideButton.setIcon(getIcon('more.png'))
            self.__showHideButton.setToolTip("Show threads list")

            self.setMinimumHeight(self.headerToolbar.height())
            self.setMaximumHeight(self.headerToolbar.height())
        else:
            self.__threadsList.setVisible(True)
            self.__showHideButton.setIcon(getIcon('less.png'))
            self.__showHideButton.setToolTip("Hide threads list")

            self.setMinimumHeight(self.__minH)
            self.setMaximumHeight(self.__maxH)
            splitterSizes = self.parent().sizes()
            splitterSizes[1] = self.splitterSize
            self.parent().setSizes(splitterSizes)

    def __resizeColumns(self):
        """Resize the files list columns"""
        self.__threadsList.header().setStretchLastSection(True)
        self.__threadsList.header().resizeSections(
            QHeaderView.ResizeToContents)
        self.__threadsList.header().resizeSection(0, 22)
        self.__threadsList.header().setSectionResizeMode(0, QHeaderView.Fixed)

    def clear(self):
        """Clears the content"""
        self.__threadsList.clear()
        self.__threadsLabel.setText("Threads")

    def populate(self, currentThreadID, threadList):
        """Populates the thread list from the client"""
        self.clear()
        for thread in threadList:
            if thread['broken']:
                state = "Waiting at breakpoint"
            else:
                state = "Running"
            item = ThreadItem(thread['id'], thread['name'], state)
            if thread['id'] == currentThreadID:
                item.setCurrent(True)
            self.__threadsList.addTopLevelItem(item)

        self.__resizeColumns()
        self.__threadsLabel.setText("Threads (total: " +
                                    str(len(threadList)) + ")")

    def switchControl(self, isInIDE):
        """Switches the UI depending where the control flow is"""
        self.__threadsList.setEnabled(isInIDE)

    # Arguments: item, column
    def __onThreadClicked(self, item, _):
        """Triggered when a thread is clicked"""
        if item.isCurrent():
            return

        for index in range(self.__threadsList.topLevelItemCount()):
            listItem = self.__threadsList.topLevelItem(index)
            if listItem.isCurrent():
                listItem.setCurrent(False)
                break
        item.setCurrent(True)

        self.__debugger.remoteSetThread(item.getTID())



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Dialog to enter a new text for a graphics item"""


from ui.qt import (Qt, QDialog, QDialogButtonBox, QVBoxLayout, QLabel,
                   QTextEdit)
from utils.colorfont import getZoomedMonoFont


class ReplaceTextDialog(QDialog):

    """Replace text input dialog"""

    def __init__(self, windowTitle, labelText, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle(windowTitle)
        self.__createLayout(labelText)

    def __createLayout(self, labelText):
        """Creates the dialog layout"""
        self.resize(600, 250)
        self.setSizeGripEnabled(True)

        # Top level layout
        layout = QVBoxLayout(self)

        layout.addWidget(QLabel(labelText))
        self.__newCaption = QTextEdit()
        self.__newCaption.setFont(getZoomedMonoFont())
        self.__newCaption.setAcceptRichText(False)
        layout.addWidget(self.__newCaption)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.close)
        layout.addWidget(buttonBox)

        self.__newCaption.setFocus()

    def setText(self, txt):
        """Sets the text to be edited"""
        self.__newCaption.setPlainText(txt)

    def text(self):
        """Provides the new text"""
        return self.__newCaption.toPlainText()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#


"""Line edit which shows specific text when inactive"""


from .qt import Qt, QLineEdit, QStyleOptionFrame, QStyle, QPainter, QPalette


class InactiveLineEdit(QLineEdit):

    """Line edit widget showing some inactive text"""

    def __init__(self, parent=None, inactiveText=""):
        QLineEdit.__init__(self, parent)
        self.__inactiveText = inactiveText

    def inactiveText(self):
        """Provides the inactive text"""
        return self.__inactiveText

    def setInactiveText(self, inactiveText):
        """Sets the inactive text"""
        self.__inactiveText = inactiveText
        self.update()

    def paintEvent(self, evt):
        """Paint event handler"""
        QLineEdit.paintEvent(self, evt)
        if self.text() == "" and self.__inactiveText != "" and \
           not self.hasFocus():
            panel = QStyleOptionFrame()
            self.initStyleOption(panel)
            textRect = self.style().subElementRect(QStyle.SE_LineEditContents,
                                                   panel, self)
            textRect.adjust(2, 0, 0, 0)
            painter = QPainter(self)
            painter.setPen(self.palette().brush(QPalette.Disabled,
                                                QPalette.Text).color())
            painter.drawText(textRect, Qt.AlignLeft | Qt.AlignVCenter,
                             self.__inactiveText)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4/6 and adopted for codimension.
# Original copyright:
# Copyright (c) 2002 - 2012 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Module implementing a debug client base class"""

import sys
import socket
import codeop
import codecs
import traceback
import os
import imp
import re
import signal
from PyQt5.QtNetwork import QTcpSocket, QAbstractSocket, QHostAddress
from protocol_cdm_dbg import (METHOD_PROC_ID_INFO, METHOD_PROLOGUE_CONTINUE,
                              METHOD_STDIN, VAR_TYPE_DISP_STRINGS,
                              METHOD_VARIABLES, METHOD_VARIABLE,
                              METHOD_THREAD_LIST, METHOD_THREAD_SET,
                              METHOD_FORK_TO,
                              METHOD_CONTINUE, METHOD_DEBUG_STARTUP,
                              METHOD_CALL_TRACE, METHOD_LINE,
                              METHOD_EXCEPTION, METHOD_STACK, METHOD_STEP_QUIT,
                              METHOD_STEP_OUT, METHOD_STEP_OVER, METHOD_STEP,
                              METHOD_MOVE_IP, METHOD_SET_BP,
                              METHOD_BP_CONDITION_ERROR, METHOD_BP_ENABLE,
                              METHOD_BP_IGNORE, METHOD_SET_WP,
                              METHOD_WP_CONDITION_ERROR, METHOD_WP_ENABLE,
                              METHOD_WP_IGNORE, METHOD_CLEAR_BP,
                              METHOD_CLEAR_WP, METHOD_SYNTAX_ERROR,
                              METHOD_SET_ENVIRONMENT, METHOD_EXECUTE_STATEMENT,
                              METHOD_SIGNAL, METHOD_SHUTDOWN,
                              METHOD_SET_FILTER, METHOD_EPILOGUE_EXIT_CODE,
                              METHOD_EPILOGUE_EXIT, SYNTAX_ERROR_AT_START,
                              STOPPED_BY_REQUEST, METHOD_EXEC_STATEMENT_ERROR,
                              METHOD_EXEC_STATEMENT_OUTPUT)
from base_cdm_dbg import setRecursionLimit
from asyncfile_cdm_dbg import AsyncFile
from outredir_cdm_dbg import OutStreamRedirector, OutStreamCollector
from bp_wp_cdm_dbg import Breakpoint, Watch
from cdm_dbg_utils import (sendJSONCommand, formatArgValues, getArgValues,
                           printerr, waitForIDEMessage,
                           getParsedJSONMessage)
from variables_cdm_dbg import getType, TOO_LARGE_ATTRIBUTE


# If set to true then the client prints debug messages on the original stderr
CLIENT_DEBUG = False

DEBUG_CLIENT_INSTANCE = None
VAR_TYPE_STRINGS = list(VAR_TYPE_DISP_STRINGS.keys())
DEBUG_CLIENT_ORIG_INPUT = None
DEBUG_CLIENT_ORIG_FORK = None
DEBUG_CLIENT_ORIG_CLOSE = None
DEBUG_CLIENT_ORIG_SET_RECURSION_LIMIT = None

WAIT_CONTINUE_TIMEOUT = 5       # in seconds
WAIT_EXIT_COMMAND_TIMEOUT = 5   # in seconds


def debugClientInput(prompt="", echo=True):
    """Replacement for the standard input builtin"""
    if DEBUG_CLIENT_INSTANCE is None or not DEBUG_CLIENT_INSTANCE.redirect:
        return DEBUG_CLIENT_ORIG_INPUT(prompt)
    return DEBUG_CLIENT_INSTANCE.input(prompt, echo)


# Use our own input().
try:
    DEBUG_CLIENT_ORIG_INPUT = __builtins__.__dict__['input']
    __builtins__.__dict__['input'] = debugClientInput
except (AttributeError, KeyError):
    import __main__
    DEBUG_CLIENT_ORIG_INPUT = __main__.__builtins__.__dict__['input']
    __main__.__builtins__.__dict__['input'] = debugClientInput


def debugClientFork():
    """Replacement for the standard os.fork()"""
    if DEBUG_CLIENT_INSTANCE is None:
        return DEBUG_CLIENT_ORIG_FORK()
    return DEBUG_CLIENT_INSTANCE.fork()


# use our own fork().
if 'fork' in dir(os):
    DEBUG_CLIENT_ORIG_FORK = os.fork
    os.fork = debugClientFork


def debugClientClose(filedesc):
    """Replacement for the standard os.close(fd)"""
    if DEBUG_CLIENT_INSTANCE is None:
        return DEBUG_CLIENT_ORIG_CLOSE(filedesc)
    return DEBUG_CLIENT_INSTANCE.close(filedesc)


# use our own close()
if 'close' in dir(os):
    DEBUG_CLIENT_ORIG_CLOSE = os.close
    os.close = debugClientClose


def debugClientSetRecursionLimit(limit):
    """Replacement for the standard sys.setrecursionlimit(limit)"""
    rlimit = max(limit, 64)
    setRecursionLimit(rlimit)
    DEBUG_CLIENT_ORIG_SET_RECURSION_LIMIT(rlimit + 64)


# use our own setrecursionlimit().
if 'setrecursionlimit' in dir(sys):
    DEBUG_CLIENT_ORIG_SET_RECURSION_LIMIT = sys.setrecursionlimit
    sys.setrecursionlimit = debugClientSetRecursionLimit
    debugClientSetRecursionLimit(sys.getrecursionlimit())


class DebugClientBase(object):

    """Class implementing the client side of the debugger.

    It provides access to the Python interpeter from a debugger running in
    another process.

    The protocol between the debugger and the client is based on JSONRPC 2.0
    PDUs. Each one is sent on a single line, i.e. commands or responses are
    separated by a linefeed character.

    If the debugger closes the session there is no response from the client.
    The client may close the session at any time as a result of the script
    being debugged closing or crashing.

    Note: This class is meant to be subclassed by individual
    DebugClient classes. Do not instantiate it directly.
    """

    # keep these in sync with VariablesViewer.VariableItem.Indicators
    INDICATORS = ['()', '[]', '{:}', '{}']

    def __init__(self):
        self.breakpoints = {}
        self.redirect = True
        self.socket = None
        self.procuuid = None

        self.__messageHandlers = {}
        self.__initMessageHandlers()

        # special objects representing the main scripts thread and frame
        self.mainThread = self
        self.framenr = 0

        # The context to run the debugged program in.
        self.debugMod = imp.new_module('__main__')
        self.debugMod.__dict__['__builtins__'] = __builtins__

        # The list of complete lines to execute.
        self.buffer = ''

        # The list of regexp objects to filter variables against
        self.globalsFilterObjects = []
        self.localsFilterObjects = []

        self._fncache = {}
        self.dircache = []
        self.passive = False    # used to indicate the passive mode
        self.running = None
        self.test = None
        self.debugging = False
        self.userInput = None

        self.forkAuto = False
        self.forkChild = False

        self.readstream = None
        self.pollingDisabled = False

        self.callTraceEnabled = None

        self.compileCommand = codeop.CommandCompiler()

        self.__encoding = 'utf-8'
        self.eventExit = None
        self.__needEpilogue = True

    def __initMessageHandlers(self):
        """initializes a map for the message handlers"""
        self.__messageHandlers = {
            METHOD_VARIABLES: self.__handleVariables,
            METHOD_VARIABLE: self.__handleVariable,
            METHOD_THREAD_LIST: self.__handleThreadList,
            METHOD_FORK_TO: self.__handleForkTo,
            METHOD_SHUTDOWN: self.__handleShutdown,
            METHOD_WP_IGNORE: self.__handleWPIgnore,
            METHOD_WP_ENABLE: self.__handleWPEnable,
            METHOD_SET_WP: self.__handleSetWP,
            METHOD_BP_IGNORE: self.__handleBPIgnore,
            METHOD_BP_ENABLE: self.__handleBPEnable,
            METHOD_SET_BP: self.__handleSetBP,
            METHOD_STDIN: self.__handleStdin,
            METHOD_CONTINUE: self.__handleContinue,
            METHOD_MOVE_IP: self.__handleMoveIP,
            METHOD_STEP_QUIT: self.__handleStepQuit,
            METHOD_STEP_OUT: self.__handleStepOut,
            METHOD_STEP_OVER: self.__handleStepOver,
            METHOD_STEP: self.__handleStep,
            METHOD_EXECUTE_STATEMENT: self.__handleExecuteStatement,
            METHOD_SET_ENVIRONMENT: self.__handleSetEnvironment,
            METHOD_CALL_TRACE: self.__handleCallTrace,
            METHOD_SET_FILTER: self.__handleSetFilter,
            METHOD_THREAD_SET: self.__handleThreadSet}

    def input(self, prompt, echo):
        """input() using the event loop"""
        sendJSONCommand(self.socket, METHOD_STDIN,
                        self.procuuid, {'prompt': prompt, 'echo': echo})
        self.eventLoop(True)
        return self.userInput

    def sessionClose(self, terminate=True):
        """Closes the session with the debugger and optionally terminate"""
        try:
            self.set_quit()
        except:
            pass

        self.debugging = False

        # make sure we close down our end of the socket
        # might be overkill as normally stdin, stdout and stderr
        # SHOULD be closed on exit, but it does not hurt to do it here
        self.readstream.close(True)

        if terminate:
            # Ok, go away.
            sys.exit()

    def __compileFileSource(self, filename, mode='exec'):
        """Compiles source code read from a file"""
        with codecs.open(filename, encoding=self.__encoding) as fp:
            statement = fp.read()

        try:
            code = compile(statement + '\n', filename, mode)
        except SyntaxError:
            exctype, excval, exctb = sys.exc_info()
            try:
                message = str(excval)
                filename = excval.filename
                lineno = excval.lineno
                charno = excval.offset
                if charno is None:
                    charno = 0
            except (AttributeError, ValueError):
                message = ""
                filename = ""
                lineno = 0
                charno = 0

            self.sendSyntaxError(message, filename, lineno, charno)
            return None
        return code

    def handleJSONCommand(self):
        """Handle a command serialized as a JSON string"""
        method, procuuid, params, _ = getParsedJSONMessage(self.socket)
        if procuuid != self.procuuid:
            return

        if CLIENT_DEBUG:
            printerr("Method: " + method + " Params: " + repr(params))

        try:
            self.__messageHandlers[method](params)
        except KeyError:
            printerr('Unhandled message. Method: ' + method +
                     ' Params: ' + repr(params))

    def __handleThreadSet(self, params):
        """Handling METHOD_THREAD_SET"""
        if params['threadID'] in self.threads:
            self.setCurrentThread(params['threadID'])
            sendJSONCommand(self.socket, METHOD_THREAD_SET,
                            self.procuuid, None)
            stack = self.currentThread.getStack()
            sendJSONCommand(self.socket, METHOD_STACK,
                            self.procuuid, {'stack': stack})

    def __handleSetFilter(self, params):
        """Handling METHOD_SET_FILTER"""
        self.__generateFilterObjects(params['scope'], params['filter'])

    def __handleCallTrace(self, params):
        """Handling METHOD_CALL_TRACE"""
        self.setCallTrace(params['enable'])

    def __handleSetEnvironment(self, params):
        """Handling METHOD_SET_ENVIRONMENT"""
        for key, value in params['environment'].items():
            if key.endswith('+'):
                if key[:-1] in os.environ:
                    os.environ[key[:-1]] += value
                else:
                    os.environ[key[:-1]] = value
            else:
                os.environ[key] = value

    def __handleExecuteStatement(self, params):
        """Handling METHOD_EXECUTE_STATEMENT"""
        statement = params['statement']
        try:
            code = self.compileCommand(statement, self.readstream.name)
        except (OverflowError, SyntaxError, ValueError):
            # Report the exception
            sys.last_type, sys.last_value, sys.last_traceback = \
                sys.exc_info()
            sendJSONCommand(self.socket, METHOD_EXEC_STATEMENT_ERROR,
                            self.procuuid,
                            {'text': ''.join(traceback.format_exception_only(
                                sys.last_type, sys.last_value))})
            return

        if code is None:
            sendJSONCommand(self.socket, METHOD_EXEC_STATEMENT_ERROR,
                            self.procuuid,
                            {'text': 'Incomplete statement to execute'})
            return

        try:
            cf = self.currentThread.getCurrentFrame()
            # program has terminated
            if cf is None:
                self.running = None
                _globals = self.debugMod.__dict__
                _locals = _globals
            else:
                frmnr = self.framenr
                while cf is not None and frmnr > 0:
                    cf = cf.f_back
                    frmnr -= 1
                _globals = cf.f_globals
                _locals = self.currentThread.getFrameLocals(self.framenr)

            # Execute a statement using a collector to collect both
            # stdout and stderr. The combined output is sent to the IDE
            collector = OutStreamCollector()
            self.__execWithCollector(code, _globals, _locals, collector)
            self.currentThread.storeFrameLocals(self.framenr)

            sendJSONCommand(self.socket, METHOD_EXEC_STATEMENT_OUTPUT,
                            self.procuuid,
                            {'text': collector.buf})

        except SystemExit as exc:
            self.progTerminated(exc.code)
        except Exception:
            # Report the exception and the traceback
            tlist = []
            try:
                exc_type, exc_value, exc_tb = sys.exc_info()
                sys.last_type = exc_type
                sys.last_value = exc_value
                sys.last_traceback = exc_tb
                tblist = traceback.extract_tb(exc_tb)
                del tblist[:1]
                tlist = traceback.format_list(tblist)
                if tlist:
                    tlist.insert(0, 'Traceback (innermost last):\n')
                    tlist.extend(traceback.format_exception_only(
                        exc_type, exc_value))
            finally:
                tblist = exc_tb = None

            sendJSONCommand(self.socket, METHOD_EXEC_STATEMENT_ERROR,
                            self.procuuid, {'text': ''.join(tlist)})

    @staticmethod
    def __execWithCollector(code, globalVars, localVars, collector):
        """The actual execution with the output collected"""
        oldStreams = [None for _ in range(6)]
        if 'sys' in globalVars:
            oldStreams[0] = globalVars['sys'].stdout
            oldStreams[1] = globalVars['sys'].stderr
            globalVars['sys'].stdout = collector
            globalVars['sys'].stderr = collector
        if 'sys' in localVars:
            oldStreams[2] = localVars['sys'].stdout
            oldStreams[3] = localVars['sys'].stderr
            localVars['sys'].stdout = collector
            localVars['sys'].stderr = collector
        oldStreams[4] = sys.stdout
        oldStreams[5] = sys.stderr
        sys.stdout = collector
        sys.stderr = collector

        try:
            exec(code, globalVars, localVars)
        finally:
            if 'sys' in globalVars:
                globalVars['sys'].stdout = oldStreams[0]
                globalVars['sys'].stderr = oldStreams[1]
            if 'sys' in localVars:
                localVars['sys'].stdout = oldStreams[2]
                localVars['sys'].stderr = oldStreams[3]
            sys.stdout = oldStreams[4]
            sys.stderr = oldStreams[5]

    def __handleStep(self, _):
        """Handling METHOD_STEP"""
        self.currentThreadExec.step(True)
        self.eventExit = True

    def __handleStepOver(self, _):
        """Handling METHOD_STEP_OVER"""
        self.currentThreadExec.step(False)
        self.eventExit = True

    def __handleStepOut(self, _):
        """Handling METHOD_STEP_OUT"""
        self.currentThreadExec.stepOut()
        self.eventExit = True

    def __handleStepQuit(self, params):
        """Handling METHOD_STEP_QUIT"""
        if self.passive:
            if params:
                if 'exitCode' in params:
                    # The IDE requested to exit with a certain code
                    self.progTerminated(params['exitCode'])
                    return
            self.progTerminated(STOPPED_BY_REQUEST)
        else:
            self.set_quit()
            self.eventExit = True

    def __handleMoveIP(self, params):
        """Handling METHOD_MOVE_IP"""
        newLine = params['newLine']
        self.currentThreadExec.move_instruction_pointer(newLine)

    def __handleContinue(self, params):
        """Handling METHOD_CONTINUE"""
        self.currentThreadExec.go(params['special'])
        self.eventExit = True

    def __handleStdin(self, params):
        """Handling METHOD_STDIN"""
        # If we are handling raw mode input then break out of the current
        # event loop.
        self.userInput = params['input']
        self.eventExit = True

    def __handleSetBP(self, params):
        """Handling METHOD_SET_BP"""
        if params['setBreakpoint']:
            if params['condition'] in ['None', '']:
                cond = None
            elif params['condition'] is not None:
                try:
                    cond = compile(params['condition'], '<string>', 'eval')
                except SyntaxError:
                    sendJSONCommand(self.socket, METHOD_BP_CONDITION_ERROR,
                                    self.procuuid,
                                    {'filename': params['filename'],
                                     'line': params['line']})
                    return
            else:
                cond = None

            Breakpoint(params['filename'], params['line'],
                       params['temporary'], cond)
        else:
            Breakpoint.clear_break(params['filename'], params['line'])

    def __handleBPEnable(self, params):
        """Handling METHOD_BP_ENABLE"""
        bPoint = Breakpoint.get_break(params['filename'], params['line'])
        if bPoint is not None:
            if params['enable']:
                bPoint.enable()
            else:
                bPoint.disable()

    def __handleBPIgnore(self, params):
        """Handling METHOD_BP_IGNORE"""
        bPoint = Breakpoint.get_break(params['filename'], params['line'])
        if bPoint is not None:
            bPoint.ignore = params['count']

    def __handleSetWP(self, params):
        """Handling METHOD_SET_WP"""
        if params['setWatch']:
            if params['condition'].endswith(
                    ('??created??', '??changed??')):
                compiledCond, flag = params['condition'].split()
            else:
                compiledCond = params['condition']
                flag = ''

            try:
                compiledCond = compile(compiledCond, '<string>', 'eval')
            except SyntaxError:
                sendJSONCommand(self.socket, METHOD_WP_CONDITION_ERROR,
                                self.procuuid,
                                {'condition': params['condition']})
                return
            Watch(params['condition'], compiledCond, flag,
                  params['temporary'])
        else:
            Watch.clear_watch(params['condition'])

    def __handleWPEnable(self, params):
        """Handling METHOD_WP_ENABLE"""
        wPoint = Watch.get_watch(params['condition'])
        if wPoint is not None:
            if params['enable']:
                wPoint.enable()
            else:
                wPoint.disable()

    def __handleWPIgnore(self, params):
        """Handling METHOD_WP_IGNORE"""
        wPoint = Watch.get_watch(params['condition'])
        if wPoint is not None:
            wPoint.ignore = params['count']

    def __handleShutdown(self, _):
        """Handling METHOD_SHUTDOWN"""
        self.sessionClose()

    def __handleForkTo(self, params):
        """Handling METHOD_FORK_TO"""
        # this results from a separate event loop
        self.forkChild = (params['target'] == 'child')
        self.eventExit = True

    def __handleVariables(self, params):
        """Handling METHOD_VARIABLES"""
        self.__dumpVariables(params['frameNumber'], params['scope'],
                             params['filters'])

    def __handleVariable(self, params):
        """Handling METHOD_VARIABLE"""
        self.__dumpVariable(params['variable'], params['frameNumber'],
                            params['scope'], params['filters'])

    def __handleThreadList(self, _):
        """Handling METHOD_THREAD_LIST"""
        self.dumpThreadList()

    def setCallTrace(self, enabled):
        """Sets up the call trace"""
        if enabled:
            sys.setprofile(self.profile)
            self.callTraceEnabled = self.profile
        else:
            sys.setprofile(None)
            self.callTraceEnabled = None

    def sendClearTemporaryBreakpoint(self, filename, lineno):
        """Signals the deletion of a temporary breakpoint"""
        sendJSONCommand(self.socket, METHOD_CLEAR_BP,
                        self.procuuid, {'filename': filename, 'line': lineno})

    def sendClearTemporaryWatch(self, condition):
        """Signals the deletion of a temporary watch expression"""
        sendJSONCommand(self.socket, METHOD_CLEAR_WP,
                        self.procuuid, {'condition': condition})

    def sendResponseLine(self, stack):
        """Sends the current call stack"""
        sendJSONCommand(self.socket, METHOD_LINE,
                        self.procuuid, {'stack': stack})

    def sendCallTrace(self, event, fromInfo, toInfo):
        """Sends a call trace entry"""
        sendJSONCommand(self.socket, METHOD_CALL_TRACE,
                        self.procuuid,
                        {'event': event[0], 'from': fromInfo, 'to': toInfo})

    def sendException(self, exceptionType, exceptionMessage, stack):
        """Sends information for an exception"""
        sendJSONCommand(self.socket, METHOD_EXCEPTION,
                        self.procuuid,
                        {'type': exceptionType, 'message': exceptionMessage,
                         'stack': stack})

    def sendSyntaxError(self, message, filename, lineno, charno):
        """Sends information for a syntax error"""
        sendJSONCommand(self.socket, METHOD_SYNTAX_ERROR,
                        self.procuuid,
                        {'message': message, 'filename': filename,
                         'line': lineno, 'characternumber': charno})

    def sendPassiveStartup(self, filename, exceptions):
        """Sends indication that the debugee is entering event loop"""
        sendJSONCommand(self.socket, METHOD_DEBUG_STARTUP,
                        self.procuuid,
                        {'filename': filename, 'exceptions': exceptions})

    def readReady(self, stream):
        """Called when there is data ready to be read"""
        self.handleJSONCommand()

    def __interact(self):
        """Interacts with the debugger"""
        global DEBUG_CLIENT_INSTANCE

        DEBUG_CLIENT_INSTANCE = self
        self.__receiveBuffer = ''

        if not self.passive:
            # At this point simulate an event loop
            self.eventLoop()

    def eventLoop(self, disablePolling=False):
        """Implements our event loop"""
        self.eventExit = None
        self.pollingDisabled = disablePolling

        while self.eventExit is None:
            if self.socket.waitForReadyRead():
                while self.socket.canReadLine():
                    self.readReady(self.readstream)

        self.eventExit = None
        self.pollingDisabled = False

    def eventPoll(self):
        """Polls for events like 'set break point'"""
        if not self.pollingDisabled:
            # without
            # if self.socket.waitForReadyRead(0):
            # canReadLine() always returns False. I have no idea why.
            # An option with
            # if self.socket.bytesAvailable() > 0:
            # does not work either. So I use a 0 timeout to make it as quick as
            # possible
            if self.socket.waitForReadyRead(0):
                while self.socket.canReadLine():
                    self.readReady(self.readstream)

    def connect(self, remoteAddress, port):
        """Establishes a session with the debugger"""
        self.socket = QTcpSocket()
        if remoteAddress is None:
            self.socket.connectToHost(QHostAddress.LocalHost, port)
        else:
            self.socket.connectToHost(remoteAddress, port)
        if not self.socket.waitForConnected(1000):
            raise Exception('Cannot connect to the IDE')
        self.socket.setSocketOption(QAbstractSocket.KeepAliveOption, 1)
        self.socket.setSocketOption(QAbstractSocket.LowDelayOption, 1)
        self.socket.disconnected.connect(self.__onDisconnected)

    def __onDisconnected(self):
        """IDE dropped the socket"""
        sys.exit(0)

    def __setupStreams(self):
        """Sets up all the required streams"""
        self.readstream = AsyncFile(self.socket, self.procuuid,
                                    sys.stdin.mode, sys.stdin.name)

        if self.redirect:
            sys.stdout = OutStreamRedirector(self.socket, True, self.procuuid)
            sys.stderr = OutStreamRedirector(self.socket, False, self.procuuid)
            sys.stdin = self.readstream

        # Attach to the main thread here
        self.attachThread(mainThread=True)

    def __unhandled_exception(self, exctype, excval, exctb):
        """Called to report an uncaught exception"""
        self.mainThread.user_exception((exctype, excval, exctb), True)

    def __interceptSignals(self):
        """Intercepts common signals"""
        for signum in [signal.SIGABRT,          # abnormal termination
                       signal.SIGFPE,           # floating point exception
                       signal.SIGILL,           # illegal instruction
                       signal.SIGSEGV]:         # segmentation violation
            signal.signal(signum, self.__signalHandler)

    def __signalHandler(self, signalNumber, stackFrame):
        """Handles signals"""
        if signalNumber == signal.SIGABRT:
            message = "Abnormal Termination"
        elif signalNumber == signal.SIGFPE:
            message = "Floating Point Exception"
        elif signalNumber == signal.SIGILL:
            message = "Illegal Instruction"
        elif signalNumber == signal.SIGSEGV:
            message = "Segmentation Violation"
        else:
            message = "Unknown Signal '{0}'".format(signalNumber)

        filename = self.absPath(stackFrame.f_code.co_filename)

        linenr = stackFrame.f_lineno
        ffunc = stackFrame.f_code.co_name

        if ffunc == '?':
            ffunc = ''

        if ffunc and not ffunc.startswith('<'):
            argInfo = getArgValues(stackFrame)
            try:
                fargs = formatArgValues(
                    argInfo.args, argInfo.varargs,
                    argInfo.keywords, argInfo.locals)
            except Exception:
                fargs = ''
        else:
            fargs = ''

        sendJSONCommand(self.socket, METHOD_SIGNAL,
                        self.procuuid,
                        {'message': message, 'filename': filename,
                         'linenumber': linenr, 'function': ffunc,
                         'arguments': fargs})

    def absPath(self, fileName):
        """Converts a filename to an absolute name"""
        if os.path.isabs(fileName):
            if sys.version_info[0] == 2:
                fileName = fileName.decode(sys.getfilesystemencoding())
            return fileName

        # Check the cache
        if fileName in self._fncache:
            return self._fncache[fileName]

        # Search sys.path
        for aPath in sys.path:
            afn = os.path.abspath(os.path.join(aPath, fileName))
            nafn = os.path.normcase(afn)

            if os.path.exists(nafn):
                if sys.version_info[0] == 2:
                    afn = afn.decode(sys.getfilesystemencoding())

                self._fncache[fileName] = afn
                aDir = os.path.dirname(afn)
                if (aDir not in sys.path) and (aDir not in self.dircache):
                    self.dircache.append(aDir)
                return afn

        # Search the additional directory cache
        for aPath in self.dircache:
            afn = os.path.abspath(os.path.join(aPath, fileName))
            nafn = os.path.normcase(afn)

            if os.path.exists(nafn):
                self._fncache[fileName] = afn
                return afn

        # Nothing found
        return fileName

    def getRunning(self):
        """True if the main script we are currently running"""
        return self.running

    def progTerminated(self, status, message=''):
        """Tells the debugger that the program has terminated"""
        if not self.__needEpilogue:
            return

        self.__needEpilogue = False

        if status is None:
            status = 0
        elif not isinstance(status, int):
            message = str(status)
            status = 1

        if self.running:
            self.set_quit()
            self.running = None

        sendJSONCommand(self.socket, METHOD_EPILOGUE_EXIT_CODE,
                        self.procuuid, {'exitCode': status,
                                        'message': message})
        waitForIDEMessage(self.socket, METHOD_EPILOGUE_EXIT,
                          WAIT_EXIT_COMMAND_TIMEOUT)
        # The final message exchange with the IDE has been completed
        # It is safe just to abort
        raise SystemExit

    def __dumpVariables(self, frmnr, scope, filterList):
        """Returns the variables of a frame to the debug server"""
        if self.currentThread is None:
            return

        frmnr += self.currentThread.skipFrames
        if scope == 0:
            self.framenr = frmnr

        f = self.currentThread.getCurrentFrame()

        while f is not None and frmnr > 0:
            f = f.f_back
            frmnr -= 1

        if f is None:
            if scope:
                varDict = self.debugMod.__dict__
            else:
                scope = -1
        elif scope:
            varDict = f.f_globals
        elif f.f_globals is f.f_locals:
            scope = -1
        else:
            varDict = f.f_locals

        varlist = []

        if scope != -1:
            keylist = varDict.keys()

            vlist = self.__formatVariablesList(
                keylist, varDict, scope, filterList)
            varlist.extend(vlist)

        sendJSONCommand(self.socket, METHOD_VARIABLES,
                        self.procuuid, {'scope': scope, 'variables': varlist})

    def __dumpVariable(self, var, frmnr, scope, filterList):
        """Returns the variables of a frame to the debug server"""
        if self.currentThread is None:
            return

        frmnr += self.currentThread.skipFrames
        f = self.currentThread.getCurrentFrame()

        while f is not None and frmnr > 0:
            f = f.f_back
            frmnr -= 1

        if f is None:
            if scope:
                varDict = self.debugMod.__dict__
            else:
                scope = -1
        elif scope:
            varDict = f.f_globals
        elif f.f_globals is f.f_locals:
            scope = -1
        else:
            varDict = f.f_locals

        varlist = []

        if scope != -1:
            variable = varDict
            for attribute in var:
                attribute = self.__extractIndicators(attribute)[0]
                typeObject, typeName, typeStr, resolver = getType(variable)
                if resolver:
                    variable = resolver.resolve(variable, attribute)
                    if variable is None:
                        break
                else:
                    break

            if variable is not None:
                typeObject, typeName, typeStr, resolver = getType(variable)
                if typeStr.startswith(('PyQt5.', 'PyQt4.')):
                    vlist = self.__formatQtVariable(variable, typeName)
                    varlist.extend(vlist)
                elif resolver:
                    varDict = resolver.getDictionary(variable)
                    vlist = self.__formatVariablesList(
                        list(varDict.keys()), varDict, scope, filterList)
                    varlist.extend(vlist)

        sendJSONCommand(self.socket, METHOD_VARIABLE,
                        self.procuuid,
                        {'scope': scope, 'variable': var,
                         'variables': varlist})

    def __extractIndicators(self, var):
        """Extracts the indicator string from a variable text"""
        for indicator in DebugClientBase.INDICATORS:
            if var.endswith(indicator):
                return var[:-len(indicator)], indicator
        return var, ""

    def __formatQtVariable(self, value, qttype):
        """Produces a formatted output of a simple Qt4/Qt5 type"""
        varlist = []
        if qttype == 'QChar':
            varlist.append(
                ('', 'QChar', '{0}'.format(chr(value.unicode()))))
            varlist.append(('', 'int', '{0:d}'.format(value.unicode())))
        elif qttype == 'QByteArray':
            varlist.append(
                ('bytes', 'QByteArray', '{0}'.format(bytes(value))[2:-1]))
            varlist.append(
                ('hex', 'QByteArray', '{0}'.format(value.toHex())[2:-1]))
            varlist.append(
                ('base64', 'QByteArray', '{0}'.format(value.toBase64())[2:-1]))
            varlist.append(('percent encoding', 'QByteArray',
                            '{0}'.format(value.toPercentEncoding())[2:-1]))
        elif qttype == 'QString':
            varlist.append(('', 'QString', '{0}'.format(value)))
        elif qttype == 'QStringList':
            for i in range(value.count()):
                varlist.append(
                    ('{0:d}'.format(i), 'QString', '{0}'.format(value[i])))
        elif qttype == 'QPoint':
            varlist.append(('x', 'int', '{0:d}'.format(value.x())))
            varlist.append(('y', 'int', '{0:d}'.format(value.y())))
        elif qttype == 'QPointF':
            varlist.append(('x', 'float', '{0:g}'.format(value.x())))
            varlist.append(('y', 'float', '{0:g}'.format(value.y())))
        elif qttype == 'QRect':
            varlist.append(('x', 'int', '{0:d}'.format(value.x())))
            varlist.append(('y', 'int', '{0:d}'.format(value.y())))
            varlist.append(('width', 'int', '{0:d}'.format(value.width())))
            varlist.append(('height', 'int', '{0:d}'.format(value.height())))
        elif qttype == 'QRectF':
            varlist.append(('x', 'float', '{0:g}'.format(value.x())))
            varlist.append(('y', 'float', '{0:g}'.format(value.y())))
            varlist.append(('width', 'float', '{0:g}'.format(value.width())))
            varlist.append(('height', 'float', '{0:g}'.format(value.height())))
        elif qttype == 'QSize':
            varlist.append(('width', 'int', '{0:d}'.format(value.width())))
            varlist.append(('height', 'int', '{0:d}'.format(value.height())))
        elif qttype == 'QSizeF':
            varlist.append(('width', 'float', '{0:g}'.format(value.width())))
            varlist.append(('height', 'float', '{0:g}'.format(value.height())))
        elif qttype == 'QColor':
            varlist.append(('name', 'str', '{0}'.format(value.name())))
            red, green, blue, alpha = value.getRgb()
            varlist.append(
                ('rgba', 'int',
                 '{0:d}, {1:d}, {2:d}, {3:d}'.format(red, green, blue, alpha)))
            hue, saturation, value, alpha = value.getHsv()
            varlist.append(
                ('hsva', 'int',
                 '{0:d}, {1:d}, {2:d}, {3:d}'.format(hue, saturation,
                                                     value, alpha)))
            cyan, magenta, yellow, black, alpha = value.getCmyk()
            varlist.append(
                ('cmyka', 'int',
                 '{0:d}, {1:d}, {2:d}, {3:d}, {4:d}'.format(cyan, magenta,
                                                            yellow, black,
                                                            alpha)))
        elif qttype == 'QDate':
            varlist.append(('', 'QDate', '{0}'.format(value.toString())))
        elif qttype == 'QTime':
            varlist.append(('', 'QTime', '{0}'.format(value.toString())))
        elif qttype == 'QDateTime':
            varlist.append(('', 'QDateTime', '{0}'.format(value.toString())))
        elif qttype == 'QDir':
            varlist.append(('path', 'str', '{0}'.format(value.path())))
            varlist.append(('absolutePath', 'str',
                            '{0}'.format(value.absolutePath())))
            varlist.append(('canonicalPath', 'str',
                            '{0}'.format(value.canonicalPath())))
        elif qttype == 'QFile':
            varlist.append(('fileName', 'str', '{0}'.format(value.fileName())))
        elif qttype == 'QFont':
            varlist.append(('family', 'str', '{0}'.format(value.family())))
            varlist.append(
                ('pointSize', 'int', '{0:d}'.format(value.pointSize())))
            varlist.append(('weight', 'int', '{0:d}'.format(value.weight())))
            varlist.append(('bold', 'bool', '{0}'.format(value.bold())))
            varlist.append(('italic', 'bool', '{0}'.format(value.italic())))
        elif qttype == 'QUrl':
            varlist.append(('url', 'str', '{0}'.format(value.toString())))
            varlist.append(('scheme', 'str', '{0}'.format(value.scheme())))
            varlist.append(('user', 'str', '{0}'.format(value.userName())))
            varlist.append(('password', 'str', '{0}'.format(value.password())))
            varlist.append(('host', 'str', '{0}'.format(value.host())))
            varlist.append(('port', 'int', '{0:d}'.format(value.port())))
            varlist.append(('path', 'str', '{0}'.format(value.path())))
        elif qttype == 'QModelIndex':
            varlist.append(('valid', 'bool', '{0}'.format(value.isValid())))
            if value.isValid():
                varlist.append(('row', 'int', '{0}'.format(value.row())))
                varlist.append(('column', 'int', '{0}'.format(value.column())))
                varlist.append(
                    ('internalId', 'int', '{0}'.format(value.internalId())))
                varlist.append(('internalPointer', 'void *',
                                '{0}'.format(value.internalPointer())))
        elif qttype == 'QRegExp':
            varlist.append(('pattern', 'str', '{0}'.format(value.pattern())))

        # GUI stuff
        elif qttype == 'QAction':
            varlist.append(('name', 'str', '{0}'.format(value.objectName())))
            varlist.append(('text', 'str', '{0}'.format(value.text())))
            varlist.append(
                ('icon text', 'str', '{0}'.format(value.iconText())))
            varlist.append(('tooltip', 'str', '{0}'.format(value.toolTip())))
            varlist.append(
                ('whatsthis', 'str', '{0}'.format(value.whatsThis())))
            varlist.append(
                ('shortcut', 'str',
                 '{0}'.format(value.shortcut().toString())))
        elif qttype == 'QKeySequence':
            varlist.append(('value', '', '{0}'.format(value.toString())))

        # XML stuff
        elif qttype == 'QDomAttr':
            varlist.append(('name', 'str', '{0}'.format(value.name())))
            varlist.append(('value', 'str', '{0}'.format(value.value())))
        elif qttype == 'QDomCharacterData':
            varlist.append(('data', 'str', '{0}'.format(value.data())))
        elif qttype == 'QDomComment':
            varlist.append(('data', 'str', '{0}'.format(value.data())))
        elif qttype == "QDomDocument":
            varlist.append(('text', 'str', '{0}'.format(value.toString())))
        elif qttype == 'QDomElement':
            varlist.append(('tagName', 'str', '{0}'.format(value.tagName())))
            varlist.append(('text', 'str', '{0}'.format(value.text())))
        elif qttype == 'QDomText':
            varlist.append(('data', 'str', '{0}'.format(value.data())))

        # Networking stuff
        elif qttype == 'QHostAddress':
            varlist.append(
                ('address', 'QHostAddress', '{0}'.format(value.toString())))

        return varlist

    def __formatVariablesList(self, keylist, dict_, scope, filterList=None,
                              formatSequences=False):
        """Produces a formated variables list"""
        filterList = [] if filterList is None else filterList[:]

        varlist = []
        if scope:
            patternFilterObjects = self.globalsFilterObjects
        else:
            patternFilterObjects = self.localsFilterObjects

        for key in keylist:
            # filter based on the filter pattern
            matched = False
            for pat in patternFilterObjects:
                if pat.match(str(key)):
                    matched = True
                    break
            if matched:
                continue

            # filter hidden attributes (filter #0)
            if 0 in filterList and str(key)[:2] == '__' and not (
                    key == "___len___" and
                    TOO_LARGE_ATTRIBUTE in keylist):
                continue

            # special handling for '__builtins__' (it's way too big)
            if key == '__builtins__':
                rvalue = '<module __builtin__ (built-in)>'
                valtype = 'module'
            else:
                value = dict_[key]
                valtypestr = str(type(value))[1:-1]
                _, valtype = valtypestr.split(' ', 1)
                valtype = valtype[1:-1]
                valtypename = type(value).__name__
                if valtype not in VAR_TYPE_STRINGS:
                    if valtype in ['numpy.ndarray', 'array.array']:
                        if VAR_TYPE_STRINGS.index('list') in filterList:
                            continue
                    elif valtypename == 'MultiValueDict':
                        if VAR_TYPE_STRINGS.index('dict') in filterList:
                            continue
                    elif valtype == 'sip.methoddescriptor':
                        if VAR_TYPE_STRINGS.index('method') in filterList:
                            continue
                    elif valtype == 'sip.enumtype':
                        if VAR_TYPE_STRINGS.index('class') in filterList:
                            continue
                    elif VAR_TYPE_STRINGS.index('instance') in filterList:
                        continue

                    if (not valtypestr.startswith('type ') and
                            valtypename not in
                            ['ndarray', 'MultiValueDict', 'array']):
                        valtype = valtypestr
                else:
                    try:
                        # Strip 'instance' to be equal with Python 3
                        if valtype == 'instancemethod':
                            valtype = 'method'

                        if VAR_TYPE_STRINGS.index(valtype) in filterList:
                            continue
                    except ValueError:
                        if valtype == 'classobj':
                            if VAR_TYPE_STRINGS.index(
                                    'instance') in filterList:
                                continue
                        elif valtype == 'sip.methoddescriptor':
                            if VAR_TYPE_STRINGS.index(
                                    'method') in filterList:
                                continue
                        elif valtype == 'sip.enumtype':
                            if VAR_TYPE_STRINGS.index('class') in \
                                    filterList:
                                continue
                        elif not valtype.startswith('PySide') and \
                            (VAR_TYPE_STRINGS.index('other') in
                             filterList):
                            continue

                try:
                    if valtype in ['list', 'tuple', 'dict', 'set',
                                   'frozenset', 'array.array']:
                        if valtype == 'dict':
                            rvalue = '{0:d}'.format(len(value.keys()))
                        else:
                            rvalue = '{0:d}'.format(len(value))
                    elif valtype == 'numpy.ndarray':
                        rvalue = '{0:d}'.format(value.size)
                    elif valtypename == 'MultiValueDict':
                        rvalue = '{0:d}'.format(len(value.keys()))
                        valtype = 'django.MultiValueDict'  # shortened type
                    else:
                        rvalue = repr(value)
                        if valtype.startswith('class') and \
                           rvalue[0] in ['{', '(', '[']:
                            rvalue = ''
                except Exception:
                    rvalue = ''

            if formatSequences:
                if str(key) == key:
                    key = "'{0!s}'".format(key)
                else:
                    key = str(key)
            varlist.append((key, valtype, rvalue))

        return varlist

    def __generateFilterObjects(self, scope, filterString):
        """Converts a filter string to a list of filter objects"""
        patternFilterObjects = []
        for pattern in filterString.split(';'):
            patternFilterObjects.append(re.compile('^{0}$'.format(pattern)))
        if scope:
            self.globalsFilterObjects = patternFilterObjects[:]
        else:
            self.localsFilterObjects = patternFilterObjects[:]

    def startProgInDebugger(self, progargs, host,
                            port, exceptions, tracePython,
                            enableCallTrace):
        """Starts the remote debugger"""
        remoteAddress = self.resolveHost(host)
        self.connect(remoteAddress, port)

        # Common part of running: the IDE waits for the client message
        sendJSONCommand(self.socket, METHOD_PROC_ID_INFO,
                        self.procuuid, None)
        waitForIDEMessage(self.socket, METHOD_PROLOGUE_CONTINUE,
                          WAIT_CONTINUE_TIMEOUT)

        self.__setupStreams()

        self._fncache = {}
        self.dircache = []
        sys.argv = progargs[:]
        sys.argv[0] = os.path.abspath(sys.argv[0])
        sys.path = self.__getSysPath(os.path.dirname(sys.argv[0]))
        self.running = sys.argv[0]
        self.debugging = True

        self.passive = True
        self.sendPassiveStartup(self.running, exceptions)
        self.__interact()

        self.attachThread(mainThread=True)
        self.mainThread.tracePythonLibs(tracePython)

        # set the system exception handling function to ensure, that
        # we report on all unhandled exceptions
        sys.excepthook = self.__unhandled_exception
        self.__interceptSignals()

        # This will eventually enter a local event loop.
        # Note the use of backquotes to cause a repr of self.running. The
        # need for this is on Windows os where backslash is the path separator.
        # They will get inadvertantly stripped away during the eval causing
        # IOErrors if self.running is passed as a normal str.
        self.debugMod.__dict__['__file__'] = self.running
        sys.modules['__main__'] = self.debugMod

        code = self.__compileFileSource(self.running)
        if code:
            self.setCallTrace(enableCallTrace)
            res = self.mainThread.run(code, self.debugMod.__dict__, debug=True)
        else:
            res = SYNTAX_ERROR_AT_START
        self.progTerminated(res)

    @staticmethod
    def resolveHost(host):
        """Resolves a hostname to an IP address"""
        try:
            host, _ = host.split("@@")
            family = socket.AF_INET6
        except ValueError:
            # version = 'v4'
            family = socket.AF_INET
        return socket.getaddrinfo(host, None, family,
                                  socket.SOCK_STREAM)[0][4][0]

    def main(self):
        """
        Public method implementing the main method.
        """
        if '--' in sys.argv:
            args = sys.argv[1:]
            host = None
            port = None
            tracePython = False
            exceptions = True
            enableCallTrace = True
            while args[0]:
                if args[0] == '--host':
                    host = args[1]
                    del args[0]
                    del args[0]
                elif args[0] == '--port':
                    port = int(args[1])
                    del args[0]
                    del args[0]
                elif args[0] == '--trace-python':
                    tracePython = True
                    del args[0]
                elif args[0] == '--no-exc-report':
                    exceptions = False
                    del args[0]
                elif args[0] == '--no-redirect':
                    self.redirect = False
                    del args[0]
                elif args[0] == '--encoding':
                    self.__encoding = args[1]
                    del args[0]
                    del args[0]
                elif args[0] == '--fork-child':
                    self.forkAuto = True
                    self.forkChild = True
                    del args[0]
                elif args[0] == '--fork-parent':
                    self.forkAuto = True
                    self.forkChild = False
                    del args[0]
                elif args[0] == '--procuuid':
                    self.procuuid = args[1]
                    del args[0]
                    del args[0]
                elif args[0] == '--no-call-trace':
                    enableCallTrace = False
                    del args[0]
                elif args[0] == '--':
                    del args[0]
                    break
                else:   # unknown option
                    del args[0]
            if not args:
                print("No program given. Aborting...")
            elif port is None or host is None:
                print("Network address is not provided. Aborting...")
            else:
                self.startProgInDebugger(args, host, port,
                                         exceptions,
                                         tracePython,
                                         enableCallTrace)
        else:
            print("No script to debug. Aborting...")

    def fork(self):
        """fork routine deciding which branch to follow"""
        # It does not make sense to follow something which was run via the
        # subprocess module. The subprocess module uses fork() internally,
        # so let's analyze it and do auto follow parent even if it was not
        # required explicitly.
        isPopen = False
        stackFrames = traceback.extract_stack()
        for stackFrame in stackFrames:
            if stackFrame[2] == '_execute_child':
                if stackFrame[0].endswith(os.path.sep + 'subprocess.py'):
                    isPopen = True

        if not self.forkAuto and not isPopen:
            sendJSONCommand(self.socket, METHOD_FORK_TO,
                            self.procuuid, None)
            self.eventLoop(True)
        pid = DEBUG_CLIENT_ORIG_FORK()

        if isPopen:
            # Switch to following parent
            oldFollow = self.forkChild
            self.forkChild = False

        if pid == 0:
            # child
            if not self.forkChild:
                sys.settrace(None)
                sys.setprofile(None)
                self.sessionClose(False)
        else:
            # parent
            if self.forkChild:
                sys.settrace(None)
                sys.setprofile(None)
                self.sessionClose(False)

        if isPopen:
            # Switch to what it was before
            self.forkChild = oldFollow
        return pid

    def close(self, fdescriptor):
        """close method as a replacement for os.close().

        It prevents the debugger connections from being closed
        """
        if fdescriptor not in [self.readstream.fileno()]:
            DEBUG_CLIENT_ORIG_CLOSE(fdescriptor)

    @staticmethod
    def __getSysPath(firstEntry):
        """calculate a path list including the PYTHONPATH env variable"""
        sysPath = [path for path in
                   os.environ.get('PYTHONPATH', '').split(os.pathsep)
                   if path not in sys.path] + sys.path[:]
        if '' in sysPath:
            sysPath.remove('')
        sysPath.insert(0, firstEntry)
        sysPath.insert(0, '')
        return sysPath



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Ignored exceptions viewer"""

from ui.qt import (Qt, QSize, QSizePolicy, QFrame, QTreeWidget, QToolButton,
                   QTreeWidgetItem, QVBoxLayout, QToolBar, QLabel, QWidget,
                   QAbstractItemView, QMenu, QHBoxLayout,
                   QCursor, QLineEdit, QPushButton, QAction)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from ui.spacers import ToolBarExpandingSpacer, ToolBarHSpacer
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.project import CodimensionProject
from utils.settings import Settings


class IgnoredExceptionsViewer(QWidget):

    """Implements the client exceptions viewer for a debugger"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.__createPopupMenu()
        self.__createLayout()
        self.__ignored = []
        self.__currentItem = None

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def __createPopupMenu(self):
        """Creates the popup menu"""
        self.__excptMenu = QMenu()
        self.__removeMenuItem = self.__excptMenu.addAction(
            getIcon('ignexcptdel.png'), "Remove from ignore list",
            self.__onRemoveFromIgnore)

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__excptLabel = HeaderFitLabel(self)
        self.__excptLabel.setText('Ignored exception types')
        self.__excptLabel.setSizePolicy(QSizePolicy.Expanding,
                                        QSizePolicy.Fixed)
        self.__excptLabel.setMinimumWidth(10)

        self.__showHideButton = QToolButton()
        self.__showHideButton.setAutoRaise(True)
        self.__showHideButton.setIcon(getIcon('less.png'))
        self.__showHideButton.setFixedSize(self.__excptLabel.height(),
                                           self.__excptLabel.height())
        self.__showHideButton.setToolTip('Hide ignored exceptions list')
        self.__showHideButton.setFocusPolicy(Qt.NoFocus)
        self.__showHideButton.clicked.connect(self.onShowHide)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(16, 16))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__excptLabel)
        self.headerToolbar.addWidget(self.__showHideButton)

        self.exceptionsList = QTreeWidget(self)
        self.exceptionsList.setSortingEnabled(False)
        self.exceptionsList.setAlternatingRowColors(True)
        self.exceptionsList.setRootIsDecorated(False)
        self.exceptionsList.setItemsExpandable(True)
        self.exceptionsList.setUniformRowHeights(True)
        self.exceptionsList.setSelectionMode(QAbstractItemView.SingleSelection)
        self.exceptionsList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.exceptionsList.setItemDelegate(NoOutlineHeightDelegate(4))
        self.exceptionsList.setContextMenuPolicy(Qt.CustomContextMenu)

        self.exceptionsList.customContextMenuRequested.connect(
            self.__showContextMenu)
        self.exceptionsList.itemSelectionChanged.connect(
            self.__onSelectionChanged)
        self.exceptionsList.setHeaderLabels(["Exception type"])

        self.__excTypeEdit = QLineEdit()
        self.__excTypeEdit.setFixedHeight(26)
        self.__excTypeEdit.textChanged.connect(self.__onNewFilterChanged)
        self.__excTypeEdit.returnPressed.connect(self.__onAddExceptionFilter)
        self.__addButton = QPushButton("Add")
        # self.__addButton.setFocusPolicy(Qt.NoFocus)
        self.__addButton.setEnabled(False)
        self.__addButton.clicked.connect(self.__onAddExceptionFilter)

        self.__removeButton = QAction(
            getIcon('delitem.png'), "Remove selected exception type", self)
        self.__removeButton.triggered.connect(self.__onRemoveFromIgnore)
        self.__removeButton.setEnabled(False)

        self.__removeAllButton = QAction(
            getIcon('ignexcptdelall.png'), "Remove all the exception types",
            self)
        self.__removeAllButton.triggered.connect(self.__onRemoveAllFromIgnore)
        self.__removeAllButton.setEnabled(False)

        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Horizontal)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedHeight(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addWidget(ToolBarExpandingSpacer(self.toolbar))
        self.toolbar.addAction(self.__removeButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addAction(self.__removeAllButton)

        addLayout = QHBoxLayout()
        addLayout.setContentsMargins(1, 1, 1, 1)
        addLayout.setSpacing(1)
        addLayout.addWidget(self.__excTypeEdit)
        addLayout.addWidget(self.__addButton)

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.toolbar)
        verticalLayout.addWidget(self.exceptionsList)
        verticalLayout.addLayout(addLayout)

    def clear(self):
        """Clears the content"""
        self.exceptionsList.clear()
        self.__excTypeEdit.clear()
        self.__addButton.setEnabled(False)
        self.__ignored = []
        self.__currentItem = None
        self.__updateTitle()

    def onShowHide(self, startup=False):
        """Triggered when show/hide button is clicked"""
        if startup or self.exceptionsList.isVisible():
            self.__minH = self.minimumHeight()
            self.__maxH = self.maximumHeight()
            self.__splitterSizes = self.parent().sizes()

            self.exceptionsList.setVisible(False)
            self.__excTypeEdit.setVisible(False)
            self.__addButton.setVisible(False)
            self.toolbar.setVisible(False)
            self.__showHideButton.setIcon(getIcon('more.png'))
            self.__showHideButton.setToolTip("Show ignored exceptions list")

            self.setMinimumHeight(self.headerToolbar.height())
            self.setMaximumHeight(self.headerToolbar.height())
        else:
            self.exceptionsList.setVisible(True)
            self.__excTypeEdit.setVisible(True)
            self.__addButton.setVisible(True)
            self.toolbar.setVisible(True)
            self.__showHideButton.setIcon(getIcon('less.png'))
            self.__showHideButton.setToolTip('Hide ignored exceptions list')

            self.setMinimumHeight(self.__minH)
            self.setMaximumHeight(self.__maxH)
            self.parent().setSizes(self.__splitterSizes)

    def __onSelectionChanged(self):
        """Triggered when the current item is changed"""
        selected = list(self.exceptionsList.selectedItems())
        if selected:
            self.__currentItem = selected[0]
            self.__removeButton.setEnabled(True)
        else:
            self.__currentItem = None
            self.__removeButton.setEnabled(False)

    def __showContextMenu(self, coord):
        """Shows the frames list context menu"""
        contextItem = self.exceptionsList.itemAt(coord)
        if contextItem is not None:
            self.__currentItem = contextItem
            self.__excptMenu.popup(QCursor.pos())

    def __updateTitle(self):
        """Updates the section title"""
        count = self.exceptionsList.topLevelItemCount()
        if count == 0:
            self.__excptLabel.setText("Ignored exception types")
        else:
            self.__excptLabel.setText("Ignored exception types (total: " +
                                      str(count) + ")")
        self.__removeAllButton.setEnabled(count != 0)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what != CodimensionProject.CompleteProject:
            return

        self.clear()
        project = GlobalData().project
        if project.isLoaded():
            self.__ignored = list(project.exceptionFilters)
        else:
            self.__ignored = Settings()['ignoredExceptions']

        for exceptionType in self.__ignored:
            item = QTreeWidgetItem(self.exceptionsList)
            item.setText(0, exceptionType)
        self.__updateTitle()

    def __onNewFilterChanged(self, text):
        """Triggered when the text is changed"""
        text = str(text).strip()
        if text == "":
            self.__addButton.setEnabled(False)
            return
        if " " in text:
            self.__addButton.setEnabled(False)
            return

        if text in self.__ignored:
            self.__addButton.setEnabled(False)
            return

        self.__addButton.setEnabled(True)

    def __onAddExceptionFilter(self):
        """Adds an item into the ignored exceptions list"""
        text = self.__excTypeEdit.text().strip()
        self.addExceptionFilter(text)

    def addExceptionFilter(self, excType):
        """Adds a new item into the ignored exceptions list"""
        if excType == "":
            return
        if " " in excType:
            return
        if excType in self.__ignored:
            return

        item = QTreeWidgetItem(self.exceptionsList)
        item.setText(0, excType)

        project = GlobalData().project
        if project.isLoaded():
            project.addExceptionFilter(excType)
        else:
            Settings().addExceptionFilter(excType)
        self.__ignored.append(excType)
        self.__updateTitle()

    def __onRemoveFromIgnore(self):
        """Removes an item from the ignored exception types list"""
        if self.__currentItem is None:
            return

        text = self.__currentItem.text(0)

        # Find the item index and remove it
        index = 0
        while True:
            if self.exceptionsList.topLevelItem(index).text(0) == text:
                self.exceptionsList.takeTopLevelItem(index)
                break
            index += 1

        project = GlobalData().project
        if project.isLoaded():
            project.deleteExceptionFilter(text)
        else:
            Settings().deleteExceptionFilter(text)
        self.__ignored.remove(text)
        self.__updateTitle()

    def __onRemoveAllFromIgnore(self):
        """Triggered when all the ignored exceptions should be deleted"""
        self.clear()

        project = GlobalData().project
        if project.isLoaded():
            project.setExceptionFilters([])
        else:
            Settings().setExceptionFilters([])

    def isIgnored(self, exceptionType):
        """Returns True if this exception type should be ignored"""
        return exceptionType in self.__ignored



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Encoding related functions"""

# pylint: disable=W0702
# pylint: disable=W0703

import re
import encodings
import logging
import os.path
from codecs import BOM_UTF8, BOM_UTF16, BOM_UTF32
from cdmpyparser import getBriefModuleInfoFromMemory
from .diskvaluesrelay import getFileEncoding
from .fileutils import isPythonFile
from .globals import GlobalData
from .settings import Settings
from .config import DEFAULT_ENCODING


# There is no way to query a complete list of the supported codecs at run-time.
# So there is the list below.
# Note: aliases are not included into the list (could be retrieved at run-time)
# Note: there could be user registered codecs as well
# Note: the list is copied from the python documentation:
#       https://docs.python.org/3/library/codecs.html
# Note: instead of the '_' char in the list the '-' was used: it looks nicer
STANDARD_CODECS = ['ascii', 'big5', 'big5hkscs', 'cp037', 'cp273', 'cp424',
                   'cp437', 'cp500', 'cp720', 'cp737', 'cp775', 'cp850',
                   'cp852', 'cp855', 'cp856', 'cp857', 'cp858', 'cp860',
                   'cp861', 'cp862', 'cp863', 'cp864', 'cp865', 'cp866',
                   'cp869', 'cp874', 'cp875', 'cp932', 'cp949', 'cp950',
                   'cp1006', 'cp1026', 'cp1125', 'cp1140', 'cp1250',
                   'cp1251', 'cp1252', 'cp1253', 'cp1254', 'cp1255',
                   'cp1256', 'cp1257', 'cp1258', 'cp65001', 'euc_jp',
                   'euc-jis-2004', 'euc-jisx0213', 'euc-kr', 'gb2312',
                   'gbk', 'gb18030', 'hz', 'iso2022-jp', 'iso2022-jp-1',
                   'iso2022-jp-2', 'iso2022-jp-2004', 'iso2022-jp-3',
                   'iso2022_jp-ext', 'iso2022-kr', 'latin-1', 'iso8859-2',
                   'iso8859-3', 'iso8859-4', 'iso8859-5', 'iso8859-6',
                   'iso8859-7', 'iso8859-8', 'iso8859-9', 'iso8859-10',
                   'iso8859-11', 'iso8859-13', 'iso8859-14', 'iso8859-15',
                   'iso8859-16', 'johab', 'koi8-r', 'koi8-t', 'koi8-u',
                   'kz1048', 'mac-cyrillic', 'mac-greek', 'mac-iceland',
                   'mac-latin2', 'mac-roman', 'mac-turkish', 'ptcp154',
                   'shift-jis', 'shift-jis-2004', 'shift-jisx0213',
                   'utf-32', 'utf-32-be', 'utf-32-le',
                   'utf-16', 'utf-16-be', 'utf-16-le',
                   'utf-7', 'utf-8', 'utf-8-sig']

# These codecs were introduced to support BOM signatures without loosing
# them in the read->modify->write cycle
SYNTHETIC_CODECS = ['bom-utf-8', 'bom-utf-16', 'bom-utf-32']

SUPPORTED_CODECS = STANDARD_CODECS + SYNTHETIC_CODECS


CODING_FROM_BYTES = [
    (2, re.compile(br'''coding[:=]\s*([-\w_.]+)''')),
    (1, re.compile(br'''<\?xml.*\bencoding\s*=\s*['"]([-\w_.]+)['"]\?>'''))]


CODING_FROM_TEXT = [
    (2, re.compile(r'''coding[:=]\s*([-\w_.]+)''')),
    (1, re.compile(r'''<\?xml.*\bencoding\s*=\s*['"]([-\w_.]+)['"]\?>'''))]


def convertLineEnds(text, eol):
    """Converts the end of line characters in text to the given eol"""
    if eol == '\r\n':
        regexp = re.compile(r"(\r(?!\n)|(?<!\r)\n)")
        return regexp.sub(lambda m, eol='\r\n': eol, text)
    if eol == '\n':
        regexp = re.compile(r"(\r\n|\r)")
        return regexp.sub(lambda m, eol='\n': eol, text)
    if eol == '\r':
        regexp = re.compile(r"(\r\n|\n)")
        return regexp.sub(lambda m, eol='\r': eol, text)
    return text


def detectEolString(text):
    """Detects the eol string using the first split. It cannot detect a mix"""
    if len(text.split('\r\n', 1)) == 2:
        return '\r\n'
    if len(text.split('\r', 1)) == 2:
        return '\r'
    return '\n'


def isValidEncoding(enc):
    """Checks if it is a valid encoding"""
    norm_enc = encodings.normalize_encoding(enc).lower()
    if norm_enc in SUPPORTED_CODECS:
        return True
    if norm_enc in [encodings.normalize_encoding(supp_enc)
                    for supp_enc in SUPPORTED_CODECS]:
        return True

    # Check the aliases as well
    if norm_enc in encodings.aliases.aliases:
        return True
    return False


def getNormalizedEncoding(enc, validityCheck=True):
    """Returns a normalized encoding or throws an exception"""
    if validityCheck:
        if not isValidEncoding(enc):
            raise Exception('Unsupported encoding ' + enc)
    norm_enc = encodings.normalize_encoding(enc).lower()
    return encodings.aliases.aliases.get(norm_enc, norm_enc)


def areEncodingsEqual(enc_lhs, enc_rhs):
    """True if the encodings are essentially the same"""
    return getNormalizedEncoding(enc_lhs) == getNormalizedEncoding(enc_rhs)


def getCodingFromBytes(text):
    """Tries to find an encoding spec from a binary file content"""
    lines = text.splitlines()
    for cfb in CODING_FROM_BYTES:
        head = lines[:cfb[0]]
        regexp = cfb[1]
        for line in head:
            match = regexp.search(line)
            if match:
                return str(match.group(1), 'ascii')
    return None


def getCodingFromText(text):
    """Tries to find an encoding spec from a text file content"""
    lines = text.splitlines()
    for cft in CODING_FROM_TEXT:
        head = lines[:cft[0]]
        regexp = cft[1]
        for line in head:
            match = regexp.search(line)
            if match:
                return match.group(1)
    return None


def encodingSanityCheck(fName, decodedText, expectedEncoding):
    """Checks if the expected encoding matches the encing in the file"""
    try:
        modInfo = getBriefModuleInfoFromMemory(decodedText)
        modEncoding = modInfo.encoding
        if modEncoding:
            if not isValidEncoding(modEncoding.name):
                logging.warning("Invalid encoding " + modEncoding.name +
                                " found in the file " + fName)
                return False
            if not areEncodingsEqual(modEncoding.name, expectedEncoding):
                if expectedEncoding.startswith('bom-'):
                    noBomEncoding = expectedEncoding[4:]
                    if areEncodingsEqual(modEncoding.name, noBomEncoding):
                        return True
                logging.warning("The explicitly set encoding " +
                                expectedEncoding +
                                " does not match encoding " + modEncoding.name +
                                " found in the file " + fName)
                return False
    except:
        pass
    return True


def detectEncodingOnClearExplicit(fName, content):
    """Provides the reading encoding as a file would be read"""
    # The function is used in case the user reset the explicit encoding
    # so the current encoding needs to be set as if the file would be
    # read again
    try:
        with open(fName, 'rb') as diskfile:
            text = diskfile.read(1024)

        if text.startswith(BOM_UTF8):
            return 'bom-utf-8'
        if text.startswith(BOM_UTF16):
            return 'bom-utf-16'
        if text.startswith(BOM_UTF32):
            return 'bom-utf-32'

        # The function is called when an explicit encoding is reset so
        # there is no need to check for it

        encFromBuffer = getCodingFromText(content)
        if encFromBuffer:
            if isValidEncoding(encFromBuffer):
                return encFromBuffer

        project = GlobalData().project
        if project.isLoaded():
            projectEncoding = project.props['encoding']
            if projectEncoding:
                if isValidEncoding(projectEncoding):
                    return projectEncoding

        ideEncoding = Settings()['encoding']
        if ideEncoding:
            if isValidEncoding(ideEncoding):
                return ideEncoding

        return DEFAULT_ENCODING
    except Exception as exc:
        logging.warning('Error while guessing encoding for reading %s: %s',
                        fName, str(exc))
        logging.warning('The default encoding %s will be used',
                        DEFAULT_ENCODING)
        return DEFAULT_ENCODING


def detectFileEncodingToRead(fName, text=None):
    """Detects the read encoding"""
    if text is None:
        with open(fName, 'rb') as diskfile:
            text = diskfile.read(1024)

    # Step 1: check for BOM
    if text.startswith(BOM_UTF8):
        return 'bom-utf-8'
    if text.startswith(BOM_UTF16):
        return 'bom-utf-16'
    if text.startswith(BOM_UTF32):
        return 'bom-utf-32'

    # Check if it was a user assigned encoding
    userAssignedEncoding = getFileEncoding(fName)
    if userAssignedEncoding:
        return userAssignedEncoding

    # Step 3: extract encoding from the file
    encFromFile = getCodingFromBytes(text)
    if encFromFile:
        return encFromFile

    # Step 4: check the project default encoding
    project = GlobalData().project
    if project.isLoaded():
        projectEncoding = project.props['encoding']
        if projectEncoding:
            return projectEncoding

    # Step 5: checks the IDE encoding
    ideEncoding = Settings()['encoding']
    if ideEncoding:
        return ideEncoding

    # Step 6: default
    return DEFAULT_ENCODING


def readEncodedFile(fName):
    """Reads the encoded file"""
    # Returns: text, used encoding
    with open(fName, 'rb') as diskfile:
        text = diskfile.read()

    isPython = isPythonFile(fName)
    triedEncodings = []
    # Step 1: check for BOM
    try:
        if text.startswith(BOM_UTF8):
            text = text[len(BOM_UTF8):]
            normEnc = encodings.normalize_encoding('utf-8')
            triedEncodings.append(normEnc)
            decodedText = str(text, normEnc)
            if isPython:
                encodingSanityCheck(fName, decodedText, 'bom-utf-8')
            return decodedText, 'bom-utf-8'
        if text.startswith(BOM_UTF16):
            text = text[len(BOM_UTF16):]
            normEnc = encodings.normalize_encoding('utf-16')
            triedEncodings.append(normEnc)
            decodedText = str(text, normEnc)
            if isPython:
                encodingSanityCheck(fName, decodedText, 'bom-utf-16')
            return decodedText, 'bom-utf-16'
        if text.startswith(BOM_UTF32):
            text = text[len(BOM_UTF32):]
            normEnc = encodings.normalize_encoding('utf-32')
            triedEncodings.append(normEnc)
            decodedText = str(text, normEnc)
            if isPython:
                encodingSanityCheck(fName, decodedText, 'bom-utf-32')
            return decodedText, 'bom-utf-32'
    except (UnicodeError, LookupError) as exc:
        logging.error('BOM signature bom-%s found in the file but decoding '
                      'failed: %s', triedEncodings[0], str(exc))
        logging.error('Continue trying...')

    # Check if it was a user assigned encoding
    userAssignedEncoding = getFileEncoding(fName)
    if userAssignedEncoding:
        if not isValidEncoding(userAssignedEncoding):
            logging.error('User assigned encoding %s is invalid. '
                          'Continue trying...', userAssignedEncoding)
        elif encodings.normalize_encoding(userAssignedEncoding) \
                not in triedEncodings:
            normEnc = encodings.normalize_encoding(userAssignedEncoding)
            triedEncodings.append(normEnc)
            try:
                decodedText = str(text, normEnc)
                if isPython:
                    encodingSanityCheck(fName, decodedText,
                                        userAssignedEncoding)
                return decodedText, userAssignedEncoding
            except (UnicodeError, LookupError) as exc:
                logging.error('Failed to decode using the user assigned '
                              'encoding %s: %s', userAssignedEncoding, str(exc))
                logging.error('Continue trying...')

    # Step 3: extract encoding from the file
    encFromFile = getCodingFromBytes(text)
    if encFromFile:
        if not isValidEncoding(encFromFile):
            logging.error('Invalid encoding found in the content: %s. '
                          'Continue trying...', encFromFile)
        elif encodings.normalize_encoding(encFromFile) not in triedEncodings:
            normEnc = encodings.normalize_encoding(encFromFile)
            triedEncodings.append(normEnc)
            try:
                decodedText = str(text, normEnc)
                if isPython:
                    encodingSanityCheck(fName, decodedText,
                                        encFromFile)
                return decodedText, encFromFile
            except (UnicodeError, LookupError) as exc:
                logging.error('Failed to decode using encoding %s found '
                              'in the file: %s', encFromFile, str(exc))
                logging.error('Continue trying...')

    # Step 4: check the project default encoding
    project = GlobalData().project
    if project.isLoaded():
        projectEncoding = project.props['encoding']
        normProjectEncoding = encodings.normalize_encoding(projectEncoding)
        if projectEncoding:
            if not isValidEncoding(projectEncoding):
                logging.error('Invalid project encoding: %s. '
                              'Continue trying...', projectEncoding)
            elif normProjectEncoding not in triedEncodings:
                triedEncodings.append(normProjectEncoding)
                try:
                    decodedText = str(text, normProjectEncoding)
                    if isPython:
                        encodingSanityCheck(fName, decodedText,
                                            projectEncoding)
                    return decodedText, projectEncoding
                except (UnicodeError, LookupError) as exc:
                    logging.error('Failed to decode using project '
                                  'encoding %s: %s', projectEncoding, str(exc))
                    logging.error('Continue trying...')

    # Step 5: checks the IDE encoding
    ideEncoding = Settings()['encoding']
    if ideEncoding:
        normIdeEnc = encodings.normalize_encoding(ideEncoding)
        if not isValidEncoding(ideEncoding):
            logging.error('Invalid ide encoding: %s. Continue trying...',
                          ideEncoding)
        elif normIdeEnc not in triedEncodings:
            triedEncodings.append(normIdeEnc)
            try:
                decodedText = str(text, normIdeEnc)
                if isPython:
                    encodingSanityCheck(fName, decodedText,
                                        ideEncoding)
                return decodedText, ideEncoding
            except (UnicodeError, LookupError) as exc:
                logging.error('Failed to decode using project encoding %s: %s',
                              ideEncoding, str(exc))
                logging.error('Continue trying...')

    # Step 6: default
    normDefEnc = encodings.normalize_encoding(DEFAULT_ENCODING)
    if normDefEnc not in triedEncodings:
        triedEncodings.append(normDefEnc)
        try:
            decodedText = str(text, normDefEnc)
            if isPython:
                encodingSanityCheck(fName, decodedText,
                                    DEFAULT_ENCODING)
            return decodedText, DEFAULT_ENCODING
        except (UnicodeError, LookupError) as exc:
            logging.error('Failed to decode using default encoding %s: %s',
                          DEFAULT_ENCODING, str(exc))
            logging.error('Continue trying...')

    # Step 7: last resort utf-8 with loosing information
    logging.warning('Last try: utf-8 decoding ignoring the errors...')
    return str(text, 'utf-8', 'ignore'), 'utf-8'


def detectNewFileWriteEncoding(editor, fName):
    """Detects a new file encoding"""
    # It could be one of two cases:
    # - the file is just created and there is no user typed content yet
    # - a new content has been modified
    isPython = isPythonFile(fName)

    if editor.explicitUserEncoding:
        # The user specifically set an encoding for a new buffer
        # It is impossible to set an invalid encoding
        if isPython:
            encFromText = getCodingFromText(editor.text)
            if encFromText:
                if not isValidEncoding(encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) is invalid and does not '
                        'match the explicitly set encoding %s. The %s is used',
                        encFromText, editor.explicitUserEncoding,
                        editor.explicitUserEncoding)
                elif not areEncodingsEqual(editor.explicitUserEncoding,
                                           encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) does not match the '
                        'explicitly set encoding %s. The %s is used',
                        encFromText, editor.explicitUserEncoding,
                        editor.explicitUserEncoding)
        return editor.explicitUserEncoding

    # This is rather paranoic. The user could have a file with a specific
    # encoding assigned. Then the file was deleted and the buffer is saved
    # again.
    userAssignedEncoding = getFileEncoding(fName)
    if userAssignedEncoding:
        if not isValidEncoding(userAssignedEncoding):
            logging.error(
                'User assigned encoding %s is invalid. Please assign a valid '
                'one and try again.', userAssignedEncoding)
            return None
        if isPython:
            encFromText = getCodingFromText(editor.text)
            if encFromText:
                if not isValidEncoding(encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) is invalid and does not '
                        'match the explicitly set encoding %s. The %s is used',
                        encFromText, userAssignedEncoding, userAssignedEncoding)
                elif not areEncodingsEqual(userAssignedEncoding,
                                           encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) does not match the '
                        'explicitly set encoding %s. The %s is used',
                        encFromText, userAssignedEncoding, userAssignedEncoding)
        return userAssignedEncoding

    # Check the buffer
    if isPython:
        encFromText = getCodingFromText(editor.text)
        if encFromText:
            if not isValidEncoding(encFromText):
                logging.error(
                    'Encoding from the buffer (%s) is invalid. Please fix the '
                    'encoding in the source or explicitly set the required one '
                    'and try again.', encFromText)
                return None
            return encFromText

    # Check the project default encoding
    project = GlobalData().project
    if project.isLoaded():
        projectEncoding = project.props['encoding']
        if projectEncoding:
            if not isValidEncoding(projectEncoding):
                logging.error(
                    'The project encoding %s is invalid. Please select a valid '
                    'one in the project properties and try again.',
                    projectEncoding)
                return None
            return projectEncoding

    # Check the IDE wide encoding
    ideEncoding = Settings()['encoding']
    if ideEncoding:
        if not isValidEncoding(ideEncoding):
            logging.error('The ide encoding %s is invalid. Please set a valid '
                          'one and try again.', ideEncoding)
            return None
        return ideEncoding

    # The default one
    return DEFAULT_ENCODING


def detectExistingFileWriteEncoding(editor, fName):
    """Provides the previously opened file encoding"""
    isPython = isPythonFile(fName)

    # The file is not new and there are a few sources of the encoding:
    # - the one which was used during reading (editor.encoding)
    # - user explicitly specified
    # - encoding in the buffer
    userAssignedEncoding = getFileEncoding(fName)
    if userAssignedEncoding:
        if not isValidEncoding(userAssignedEncoding):
            logging.error(
                'User assigned encoding %s is invalid. Please assign a valid '
                'one and try again.', userAssignedEncoding)
            return None
        if isPython:
            encFromText = getCodingFromText(editor.text)
            if encFromText:
                if not isValidEncoding(encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) is invalid and '
                        'does not match the explicitly set encoding %s. '
                        'The %s is used',
                        encFromText, userAssignedEncoding, userAssignedEncoding)
                elif not areEncodingsEqual(userAssignedEncoding,
                                           encFromText):
                    logging.warning(
                        'Encoding from the buffer (%s) does not match the '
                        'explicitly set encoding %s. The %s is used',
                        encFromText, userAssignedEncoding, userAssignedEncoding)
        return userAssignedEncoding

    # Check the buffer
    if isPython:
        encFromText = getCodingFromText(editor.text)
        if encFromText:
            if not isValidEncoding(encFromText):
                logging.error(
                    'Encoding from the buffer (%s) is invalid. Please fix the'
                    'encoding in the source or explicitly set the required one '
                    'and try again.', encFromText)
                return None
            return encFromText

    # Here: no explicitly specified encoding, no encoding in the buffer,
    #       then use the encoding the file was read with
    return editor.encoding


def detectWriteEncoding(editor, fName):
    """Detects the write encoding for a buffer"""
    # If editor.encoding is None => the file has never been saved
    # At the same time fName may exist, i.e. a new file overwrites the existing
    # one.
    if os.path.isabs(fName) and os.path.exists(fName) and \
        editor.encoding is not None:
        return detectExistingFileWriteEncoding(editor, fName)
    return detectNewFileWriteEncoding(editor, fName)


def writeEncodedFile(fName, content, encoding):
    """Writes into a file taking care of encoding"""
    normEnc = getNormalizedEncoding(encoding)
    try:
        if normEnc.startswith('bom_'):
            enc = normEnc[4:]
            if enc == 'utf_8':
                encContent = BOM_UTF8 + content.encode(enc)
            elif enc == 'utf_16':
                encContent = BOM_UTF16 + content.encode(enc)
            else:
                encContent = BOM_UTF32 + content.encode(enc)
        else:
            encContent = content.encode(normEnc)

            # Workaround for empty files: if there is no visible content and
            # the file is saved then the editor reports precisely \n which is
            # saved on disk and then detected as octet-stream. If there are
            # more than one \n then the file is detected as plain text.
            # The octet stream files are not openable in Codimension
            if encContent == b'\n':
                encContent = b''
    except (UnicodeError, LookupError) as exc:
        raise Exception('Error encoding the buffer content with ' + encoding +
                        ': ' + str(exc))

    try:
        with open(fName, 'wb') as diskfile:
            diskfile.write(encContent)
    except Exception as exc:
        raise Exception('Error writing encoded buffer content into ' +
                        fName + ': ' + str(exc))


def decodeURLContent(content):
    """Decodes the content read from a URL"""
    project = GlobalData().project
    if project.isLoaded():
        projectEncoding = project.props['encoding']
        if projectEncoding:
            if not isValidEncoding(projectEncoding):
                raise Exception(
                    "The prject encoding " + projectEncoding + " is invalid. "
                    "Please select a valid one in the project properties and "
                    "try again.")
            return content.decode(
                encodings.normalize_encoding(projectEncoding))

    # Check the IDE wide encoding
    ideEncoding = Settings()['encoding']
    if ideEncoding:
        if not isValidEncoding(ideEncoding):
            raise Exception("The ide encoding " + ideEncoding + " is invalid. "
                            "Please set a valid one and try again.")
        return content.decode(encodings.normalize_encoding(ideEncoding))

    # The default one
    return content.decode(DEFAULT_ENCODING)



#!/bin/env python
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2011-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The idea for the calculating the valid lines is taken from rpdb2.
#

"""Breakpoint related utilities"""

from utils.fileutils import getFileContent

validBreakPointLinesCache = {}


def getBreakpointLine(userLine, isEmptyLine, srcCode):
    """Provides the line where the debugger can actually break"""
    # The function should be called for a python buffer only
    try:
        validLines = calcBreakpointLines(srcCode)
        if validLines is None:
            # If there is a compile error, let the user decide
            return userLine
    except:
        # If there is a compile error, let the user decide
        return userLine

    if userLine in validLines:
        return userLine

    # Not a valid breakpoint line. Search the closest suitable.
    validLines = list(validLines).sort()
    if isEmptyLine:
        # Take a line after this one
        for line in validLines:
            if line > userLine:
                return line
        # There are no lines after the given one
        if validLines:
            return validLines[-1]
        # There are no valid lines at all
        return None

    # Take a line before this one
    for index in range(len(validLines), -1, -1):
        if validLines[index] < userLine:
            return validLines[index]
    # There are no lines after the given one
    if validLines:
        return validLines[0]
    # There are no valid lines at all
    return None


def clearValidBreakpointLinesCache():
    """Resets the cache"""
    global validBreakPointLinesCache
    validBreakPointLinesCache = {}


def getBreakpointLines(fileName, srcCode,
                       enforceRecalc=False, saveToCache=True):
    """Provides a set of breakable lines"""
    global validBreakPointLinesCache
    if not enforceRecalc:
        if fileName in validBreakPointLinesCache:
            return validBreakPointLinesCache[fileName]

    try:
        if srcCode is None:
            srcCode = getFileContent(fileName)
        lines = calcBreakpointLines(srcCode)
        if saveToCache:
            validBreakPointLinesCache[fileName] = lines
        return lines
    except:
        return None


def calcBreakpointLines(sourceCode):
    """Calculates valid breakpoint lines"""
    def __safeOrd(char):
        """Exception safe ord"""
        try:
            return ord(char)
        except:
            return char

    def __calcValidLines(code, validLines):
        """Calculates valid breakpoint lines"""
        l = code.co_firstlineno
        validLines.add(l)
        bl = [__safeOrd(c) for c in code.co_lnotab[2::2]]
        sl = [__safeOrd(c) for c in code.co_lnotab[1::2]]
        for (bi, si) in zip(bl, sl):
            l += si
            if bi == 0:
                continue
            validLines.add(l)
        if sl:
            l += sl[-1]
            validLines.add(l)

    def __calcSubCodesList(code):
        """Adds nested fragments"""
        tc = type(code)
        t = [(c.co_firstlineno, c) for c in code.co_consts if type(c) == tc]
        t.sort()
        scl = [c[1] for c in t]
        return scl

    code = compile(sourceCode, '', "exec")

    t = [code]
    validLines = set()

    while t:
        c = t.pop(0)
        __calcValidLines(c, validLines)
        subcodeslist = __calcSubCodesList(c)
        t = subcodeslist + t
    return validLines


if __name__ == "__main__":
    import sys
    if len(sys.argv) != 2:
        print("Expected python file name", file=sys.stderr)
        sys.exit(1)

    print("Valid break point lines:")
    print(calcBreakpointLines(open(sys.argv[1]).read()))
    sys.exit(0)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Color mixin for the graphics items"""

# pylint: disable=C0305

from .cml import CMLVersion, CMLcc

# There are a few options of how the colors could be specified:
# - the item may have a leading CML cc comment
# - a scope item has a docstring which in turn has a leading CML cc comment
#   a docstring color spec does not support a border color because the scope
#   border is used for the dividing line
# - group items have the color spec in their CML ref
# - doc items have the color spec in their CML ref

class ColorMixin:

    """Color mixin to support bg, fg and border color"""

    def __init__(self, ref, defaultBG, defaultFG, defaultBorder,
                 isDocstring=False, colorSpec=None):
        self.bgColor = defaultBG
        self.fgColor = defaultFG
        self.borderColor = defaultBorder

        if colorSpec is not None:
            # the case for groups and docs
            self.__getFromColorSpec(colorSpec)
        elif isDocstring:
            # special case for docstrings
            self.__getFromDocstring(ref)
        else:
            # all the other items
            self.__getCustomColors(ref)

    def getColors(self):
        """Provides the item colors"""
        return self.bgColor, self.fgColor, self.borderColor

    @staticmethod
    def isDark(red, green, blue):
        """True if the color is dark"""
        yiq = ((red * 299) + (green * 587) + (blue * 114)) / 1000
        return yiq < 128

    def __getFromColorSpec(self, colorSpec):
        """Updates the colors from the given colorspec"""
        if colorSpec.bgColor:
            self.bgColor = colorSpec.bgColor
        if colorSpec.fgColor:
            self.fgColor = colorSpec.fgColor
        if colorSpec.border:
            self.borderColor = colorSpec.border

    def __getFromDocstring(self, ref):
        """Updates the colors for the docstrings"""
        leadingCML = ref.docstring.leadingCMLComments
        if leadingCML:
            colorSpec = CMLVersion.find(leadingCML, CMLcc)
            if colorSpec:

                # NOTE: no border color support for the docstrings

                if colorSpec.bgColor:
                    self.bgColor = colorSpec.bgColor
                if colorSpec.fgColor:
                    self.fgColor = colorSpec.fgColor

    def __getCustomColors(self, ref):
        """Provides the colors to be used for an item"""
        leadingCML = ref.leadingCMLComments
        if leadingCML:
            colorSpec = CMLVersion.find(leadingCML, CMLcc)
            if colorSpec:
                self.__getFromColorSpec(colorSpec)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""New nested project directory dialog"""

import os
from .qt import (Qt, QEvent, QObject, QDialog, QLineEdit, QLabel,
                 QDialogButtonBox, QVBoxLayout)


class NewProjectDirDialog(QDialog):

    """New project directory dialog"""

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)

        self.__dirnameEdit = None
        self.__buttonBox = None
        self.__createLayout()

        self.okButton = self.__buttonBox.button(QDialogButtonBox.Ok)
        self.okButton.setEnabled(False)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(400, 100)
        self.setWindowTitle("Create subdirectory")
        vboxlayout = QVBoxLayout(self)

        inputLabel = QLabel(self)
        inputLabel.setText("Type new subdirectory name")
        vboxlayout.addWidget(inputLabel)

        self.__dirnameEdit = QLineEdit(self)
        self.__dirnameEdit.setToolTip("Subdirectory name without "
                                      "path separators")
        self.__dirnameEdit.installEventFilter(self)
        self.__dirnameEdit.textEdited.connect(self.__onTextChanged)
        vboxlayout.addWidget(self.__dirnameEdit)

        self.__buttonBox = QDialogButtonBox(self)
        self.__buttonBox.setOrientation(Qt.Horizontal)
        self.__buttonBox.setStandardButtons(QDialogButtonBox.Cancel |
                                            QDialogButtonBox.Ok)
        vboxlayout.addWidget(self.__buttonBox)

        self.__buttonBox.accepted.connect(self.accept)
        self.__buttonBox.rejected.connect(self.reject)

    def __onTextChanged(self, text):
        """Triggered when the input text has been changed"""
        self.okButton.setEnabled(text != '')

    def getDirName(self):
        """Provides the user input"""
        return self.__dirnameEdit.text()

    def eventFilter(self, obj, event):
        """Event filter for the project name field"""
        # Do not allow path separators
        if event.type() == QEvent.KeyPress:
            if event.key() == ord(os.path.sep):
                return True
        return QObject.eventFilter(self, obj, event)
